import os
import pathlib
import warnings

import skimage.io
from multiprocess import Pool
from functools import partial

import numpy as np
from pycocotools.coco import COCO
import shapely.geometry

from tqdm import tqdm

import torch

from lydorn_utils import print_utils
from lydorn_utils import python_utils

from torch_lydorn.torch.utils.data import Dataset as LydornDataset, makedirs, __repr__

from torch_lydorn.torchvision.datasets import utils


class MappingChallenge(LydornDataset):
    def __init__(self, root, transform=None, pre_transform=None, fold="train", small=False, pool_size=1):
        assert fold in ["train", "val", "test_images"], "Input fold={} should be in [\"train\", \"val\", \"test_images\"]".format(fold)
        if fold == "test_images":
            print_utils.print_error("ERROR: fold {} not yet implemented!".format(fold))
            exit()
        self.root = root
        self.fold = fold
        makedirs(self.processed_dir)
        self.small = small
        if self.small:
            print_utils.print_info("INFO: Using small version of the Mapping challenge dataset.")
        self.pool_size = pool_size

        self.coco = None
        self.image_id_list = self.load_image_ids()
        self.stats_filepath = os.path.join(self.processed_dir, "stats.pt")
        self.stats = None
        if os.path.exists(self.stats_filepath):
            self.stats = torch.load(self.stats_filepath)
        self.processed_flag_filepath = os.path.join(self.processed_dir, "processed-flag-small" if self.small else "processed-flag")

        super(MappingChallenge, self).__init__(root, transform, pre_transform)

    def load_image_ids(self):
        image_id_list_filepath = os.path.join(self.processed_dir, "image_id_list-small.json" if self.small else "image_id_list.json")
        if os.path.exists(image_id_list_filepath):
            image_id_list = python_utils.load_json(image_id_list_filepath)
        else:
            coco = self.get_coco()
            image_id_list = coco.getImgIds(catIds=coco.getCatIds())
        # Save for later so that the whole coco object doesn't have to be instantiated when just reading processed samples with multiple workers:
        python_utils.save_json(image_id_list_filepath, image_id_list)
        return image_id_list

    def get_coco(self):
        if self.coco is None:
            annotation_filename = "annotation-small.json" if self.small else "annotation.json"
            annotations_filepath = os.path.join(self.raw_dir, self.fold, annotation_filename)
            self.coco = COCO(annotations_filepath)
        return self.coco

    @property
    def processed_dir(self):
        return os.path.join(self.root, 'processed', self.fold)

    @property
    def processed_file_names(self):
        ell = []
        for image_id in self.image_id_list:
            ell.append(os.path.join("data_{:012d}.pt".format(image_id)))
        return ell

    def __len__(self):
        return len(self.image_id_list)

    def _download(self):
        pass

    def download(self):
        pass

    def _process(self):
        f = os.path.join(self.processed_dir, 'pre_transform.pt')
        if os.path.exists(f) and torch.load(f) != __repr__(self.pre_transform):
            warnings.warn(
                'The `pre_transform` argument differs from the one used in '
                'the pre-processed version of this dataset. If you really '
                'want to make use of another pre-processing technique, make '
                'sure to delete `{}` first.'.format(self.processed_dir))
        f = os.path.join(self.processed_dir, 'pre_filter.pt')
        if os.path.exists(f) and torch.load(f) != __repr__(self.pre_filter):
            warnings.warn(
                'The `pre_filter` argument differs from the one used in the '
                'pre-processed version of this dataset. If you really want to '
                'make use of another pre-fitering technique, make sure to '
                'delete `{}` first.'.format(self.processed_dir))

        if os.path.exists(self.processed_flag_filepath):
            return

        print('Processing...')

        makedirs(self.processed_dir)
        self.process()

        path = os.path.join(self.processed_dir, 'pre_transform.pt')
        torch.save(__repr__(self.pre_transform), path)
        path = os.path.join(self.processed_dir, 'pre_filter.pt')
        torch.save(__repr__(self.pre_filter), path)

        print('Done!')

    def process(self):
        images_relative_dirpath = os.path.join("raw", self.fold, "images")

        image_info_list = []
        coco = self.get_coco()
        for image_id in self.image_id_list:
            filename = coco.loadImgs(image_id)[0]["file_name"]
            annotation_ids = coco.getAnnIds(imgIds=image_id)
            annotation_list = coco.loadAnns(annotation_ids)
            image_info = {
                "image_id": image_id,
                "image_filepath": os.path.join(self.root, images_relative_dirpath, filename),
                "image_relative_filepath": os.path.join(images_relative_dirpath, filename),
                "annotation_list": annotation_list
            }
            image_info_list.append(image_info)

        partial_preprocess_one = partial(preprocess_one, pre_filter=self.pre_filter, pre_transform=self.pre_transform,
                                         processed_dir=self.processed_dir)
        with Pool(self.pool_size) as p:
            sample_stats_list = list(tqdm(p.imap(partial_preprocess_one, image_info_list), total=len(image_info_list)))

        # Aggregate sample_stats_list
        image_s0_list, image_s1_list, image_s2_list, class_freq_list = zip(*sample_stats_list)
        image_s0_array = np.stack(image_s0_list, axis=0)
        image_s1_array = np.stack(image_s1_list, axis=0)
        image_s2_array = np.stack(image_s2_list, axis=0)
        class_freq_array = np.stack(class_freq_list, axis=0)

        image_s0_total = np.sum(image_s0_array, axis=0)
        image_s1_total = np.sum(image_s1_array, axis=0)
        image_s2_total = np.sum(image_s2_array, axis=0)

        image_mean = image_s1_total / image_s0_total
        image_std = np.sqrt(image_s2_total / image_s0_total - np.power(image_mean, 2))
        class_freq = np.sum(class_freq_array * image_s0_array[:, None], axis=0) / image_s0_total

        # Save aggregated stats
        self.stats = {
            "image_mean": image_mean,
            "image_std": image_std,
            "class_freq": class_freq,
        }
        torch.save(self.stats, self.stats_filepath)

        # Indicates that processing has been performed:
        pathlib.Path(self.processed_flag_filepath).touch()

    def get(self, idx):
        image_id = self.image_id_list[idx]
        data = torch.load(os.path.join(self.processed_dir, "data_{:012d}.pt".format(image_id)))
        data["image_mean"] = self.stats["image_mean"]
        data["image_std"] = self.stats["image_std"]
        data["class_freq"] = self.stats["class_freq"]
        return data


def preprocess_one(image_info, pre_filter, pre_transform, processed_dir):
    out_filepath = os.path.join(processed_dir, "data_{:012d}.pt".format(image_info["image_id"]))
    data = None
    if os.path.exists(out_filepath):
        # Load already-processed sample
        try:
            data = torch.load(out_filepath)
        except EOFError:
            pass
    if data is None:
        # Process sample:
        image = skimage.io.imread(image_info["image_filepath"])
        gt_polygons = []
        for annotation in image_info["annotation_list"]:
            flattened_segmentation_list = annotation["segmentation"]
            if len(flattened_segmentation_list) != 1:
                print("WHAT!?!, len(flattened_segmentation_list = {}".format(len(flattened_segmentation_list)))
                print("To implement: if more than one segmentation in flattened_segmentation_list (MS COCO format), does it mean it is a MultiPolygon or a Polygon with holes?")
                raise NotImplementedError
            flattened_arrays = np.array(flattened_segmentation_list)
            coords = np.reshape(flattened_arrays, (-1, 2))
            polygon = shapely.geometry.Polygon(coords)

            # Filter out degenerate polygons (area is lower than 2.0)
            if 2.0 < polygon.area:
                gt_polygons.append(polygon)

        data = {
            "image": image,
            "gt_polygons": gt_polygons,
            "image_relative_filepath": image_info["image_relative_filepath"],
            "name": os.path.splitext(os.path.basename(image_info["image_relative_filepath"]))[0],
            "image_id": image_info["image_id"]
        }

        if pre_filter is not None and not pre_filter(data):
            return

        if pre_transform is not None:
            data = pre_transform(data)

        # masked_angles = data["gt_crossfield_angle"].astype(np.float) * data["gt_polygons_image"][:, :, 1].astype(np.float)
        # skimage.io.imsave("gt_crossfield_angle.png", data["gt_crossfield_angle"])
        # skimage.io.imsave("masked_angles.png", masked_angles)
        # exit()

        torch.save(data, out_filepath)

    # Compute stats for later aggregation for the whole dataset
    normed_image = data["image"] / 255
    image_s0 = data["image"].shape[0] * data["image"].shape[1]  # Number of pixels
    image_s1 = np.sum(normed_image, axis=(0, 1))  # Sum of pixel normalized values
    image_s2 = np.sum(np.power(normed_image, 2), axis=(0, 1))
    class_freq = np.mean(data["gt_polygons_image"], axis=(0, 1)) / 255

    return image_s0, image_s1, image_s2, class_freq


def main():
    # Test using transforms from the frame_field_learning project:
    from frame_field_learning import data_transforms

    config = {
        "data_dir_candidates": [
                "/data/titane/user/nigirard/data",
                "~/data",
                "/data"
        ],
        "dataset_params": {
            "small": True,
            "root_dirname": "mapping_challenge_dataset",
            "seed": 0,
            "train_fraction": 0.75
        },
        "num_workers": 8,
        "data_aug_params": {
            "enable": False,
            "vflip": True,
            "affine": True,
            "color_jitter": True,
            "device": "cuda"
        }
    }

    # Find data_dir
    data_dir = python_utils.choose_first_existing_path(config["data_dir_candidates"])
    if data_dir is None:
        print_utils.print_error("ERROR: Data directory not found!")
        exit()
    else:
        print_utils.print_info("Using data from {}".format(data_dir))
    root_dir = os.path.join(data_dir, config["dataset_params"]["root_dirname"])

    # --- Transforms: --- #
    # --- pre-processing transform (done once then saved on disk):
    # --- Online transform done on the host (CPU):
    train_online_cpu_transform = data_transforms.get_online_cpu_transform(config,  # NOQA
                                                                          augmentations=config["data_aug_params"][
                                                                              "enable"])
    test_online_cpu_transform = data_transforms.get_eval_online_cpu_transform()

    train_online_cuda_transform = data_transforms.get_online_cuda_transform(config,
                                                                            augmentations=config["data_aug_params"][
                                                                                "enable"])
    # --- --- #

    dataset = MappingChallenge(root_dir,
                               transform=test_online_cpu_transform,
                               pre_transform=data_transforms.get_offline_transform_patch(),
                               fold="train",
                               small=config["dataset_params"]["small"],
                               pool_size=config["num_workers"])

    print("# --- Sample 0 --- #")
    sample = dataset[0]
    print(sample.keys())

    for key, item in sample.items():
        print("{}: {}".format(key, type(item)))

    print(sample["image"].shape)
    print(len(sample["gt_polygons_image"]))
    print("# --- Samples --- #")
    # for data in tqdm(dataset):
    #     pass

    data_loader = torch.utils.data.DataLoader(dataset, batch_size=10, shuffle=True, num_workers=config["num_workers"])
    print("# --- Batches --- #")
    for batch in tqdm(data_loader):
        print("Images:")
        print(batch["image_relative_filepath"])
        print(batch["image"].shape)
        print(batch["gt_polygons_image"].shape)

        print("Apply online tranform:")
        batch = utils.batch_to_cuda(batch)
        batch = train_online_cuda_transform(batch)
        batch = utils.batch_to_cpu(batch)

        print(batch["image"].shape)
        print(batch["gt_polygons_image"].shape)

        # Save output to visualize
        seg = np.array(batch["gt_polygons_image"][0])
        seg = np.moveaxis(seg, 0, -1)
        seg_display = utils.get_seg_display(seg)
        seg_display = (seg_display * 255).astype(np.uint8)
        skimage.io.imsave("gt_seg.png", seg_display)
        skimage.io.imsave("gt_seg_edge.png", seg[:, :, 1])

        im = np.array(batch["image"][0])
        im = np.moveaxis(im, 0, -1)
        skimage.io.imsave('im.png', im)

        gt_crossfield_angle = np.array(batch["gt_crossfield_angle"][0])
        gt_crossfield_angle = np.moveaxis(gt_crossfield_angle, 0, -1)
        skimage.io.imsave('gt_crossfield_angle.png', gt_crossfield_angle)

        distances = np.array(batch["distances"][0])
        distances = np.moveaxis(distances, 0, -1)
        skimage.io.imsave('distances.png', distances)

        sizes = np.array(batch["sizes"][0])
        sizes = np.moveaxis(sizes, 0, -1)
        skimage.io.imsave('sizes.png', sizes)

        # valid_mask = np.array(batch["valid_mask"][0])
        # valid_mask = np.moveaxis(valid_mask, 0, -1)
        # skimage.io.imsave('valid_mask.png', valid_mask)

        input("Press enter to continue...")


if __name__ == '__main__':
    main()



import numpy as np
import time
import json
import os.path
from tqdm import tqdm

import rasterio
from osgeo import gdal, ogr
from osgeo import osr
import overpy
from pyproj import Proj, transform
import fiona
import fiona.crs
import shapely.geometry
import shapely.ops

from . import polygon_utils
from . import math_utils
from . import print_utils

# --- Params --- #

QUERY_BASE = \
    """
    <osm-script timeout="900" element-limit="1073741824">
      <union>
        <query type="way">
          <has-kv k="{0}"/>
          <bbox-query s="{1}" w="{2}" n="{3}" e="{4}"/>
        </query>
        <recurse type="way-node" into="nodes"/>
      </union>
      <print/>
    </osm-script>
    """

WGS84_WKT = """
    GEOGCS["GCS_WGS_1984",
        DATUM["WGS_1984",
            SPHEROID["WGS_84",6378137,298.257223563]],
                PRIMEM["Greenwich",0],
        UNIT["Degree",0.017453292519943295]]
        """

CRS = {'no_defs': True, 'ellps': 'WGS84', 'datum': 'WGS84', 'proj': 'longlat'}


# --- --- #


def get_coor_in_space(image_filepath):
    """

    :param image_filepath: Path to geo-referenced tif image
    :return: coor in original space and in wsg84 spatial reference and original geotransform
    :return: geo transform (x_min, res, 0, y_max, 0, -res)
    :return: [[OR_x_min,OR_y_min,OR_x_max,OR_y_max],[TR_x_min,TR_y_min,TR_x_max,TR_y_max]]
    """
    # print(" get_coor_in_space(image_filepath)")
    ds = gdal.Open(image_filepath)
    width = ds.RasterXSize
    height = ds.RasterYSize
    gt = ds.GetGeoTransform()

    x_min = gt[0]
    y_min = gt[3] + width * gt[4] + height * gt[5]
    x_max = gt[0] + width * gt[1] + height * gt[2]
    y_max = gt[3]

    prj = ds.GetProjection()
    srs = osr.SpatialReference(wkt=prj)

    coor_sys = srs.GetAttrValue("PROJCS|AUTHORITY", 1)

    if coor_sys is None:
        coor_sys = srs.GetAttrValue("GEOGCS|AUTHORITY", 1)

    new_cs = osr.SpatialReference()
    new_cs.ImportFromWkt(WGS84_WKT)

    # print(srs, new_cs)
    transform = osr.CoordinateTransformation(srs, new_cs)

    lat_long_min = transform.TransformPoint(x_min, y_min)
    lat_long_max = transform.TransformPoint(x_max, y_max)

    coor = [[x_min, y_min, x_max, y_max], [lat_long_min[0], lat_long_min[1], lat_long_max[0], lat_long_max[1]]]
    return coor, gt, coor_sys


def get_osm_data(coor_query):
    """

    :param coor_query: [x_min, min_z, x_max, y_max]
    :return: OSM query result
    """
    api = overpy.Overpass()
    query_buildings = QUERY_BASE.format("building", coor_query[1], coor_query[0], coor_query[3], coor_query[2])
    query_successful = False
    wait_duration = 60
    result = None
    while not query_successful:
        try:
            result = api.query(query_buildings)
            query_successful = True
        except overpy.exception.OverpassGatewayTimeout or overpy.exception.OverpassTooManyRequests or ConnectionResetError:
            print("OSM server overload. Waiting for {} seconds before querying again...".format(wait_duration))
            time.sleep(wait_duration)
            wait_duration *= 2  # Multiply wait time by 2 for the next time
    return result


def proj_to_epsg_space(nodes, coor_sys):
    original = Proj(CRS)
    destination = Proj(init='EPSG:{}'.format(coor_sys))
    polygon = []
    for node in nodes:
        polygon.append(transform(original, destination, node.lon, node.lat))
    return np.array(polygon)


def compute_epsg_to_image_mat(coor, gt):
    x_min = coor[0][0]
    y_max = coor[0][3]

    transform_mat = np.array([
        [gt[1], 0, 0],
        [0, gt[5], 0],
        [x_min, y_max, 1],
    ])
    return np.linalg.inv(transform_mat)


def compute_image_to_epsg_mat(coor, gt):
    x_min = coor[0][0]
    y_max = coor[0][3]

    transform_mat = np.array([
        [gt[1], 0, 0],
        [0, gt[5], 0],
        [x_min, y_max, 1],
    ])
    return transform_mat


def apply_transform_mat(polygon_epsg_space, transform_mat):
    polygon_epsg_space_homogeneous = math_utils.to_homogeneous(polygon_epsg_space)
    polygon_image_space_homogeneous = np.matmul(polygon_epsg_space_homogeneous, transform_mat)
    polygon_image_space = math_utils.to_euclidian(polygon_image_space_homogeneous)
    return polygon_image_space


def get_polygons_from_osm(image_filepath, tag="", ij_coords=True):
    coor, gt, coor_system = get_coor_in_space(image_filepath)
    transform_mat = compute_epsg_to_image_mat(coor, gt)
    osm_data = get_osm_data(coor[1])

    polygons = []
    for way in osm_data.ways:
        if way.tags.get(tag, "n/a") != 'n/a':
            polygon = way.nodes
            polygon_epsg_space = proj_to_epsg_space(polygon, coor_system)
            polygon_image_space = apply_transform_mat(polygon_epsg_space, transform_mat)
            if ij_coords:
                polygon_image_space = polygon_utils.swap_coords(polygon_image_space)
            polygons.append(polygon_image_space)

    return polygons


def get_polygons_from_shapefile(image_filepath, input_shapefile_filepath):
    def process_one_polygon(polygon):
        assert len(polygon.shape) == 2, "polygon should have shape (n, d), not {}".format(polygon.shape)
        if 2 < polygon.shape[1]:
            print_utils.print_warning(
                "WARNING: polygon from shapefile has shape {}. Will discard extra values to have polygon with shape ({}, 2)".format(
                    polygon.shape, polygon.shape[0]))
            polygon = polygon[:, :2]
        polygon_epsg_space = polygon
        polygon_image_space = apply_transform_mat(polygon_epsg_space, transform_mat)
        polygon_image_space = polygon_utils.swap_coords(polygon_image_space)
        polygons.append(polygon_image_space)

        # Extract properties:
        if "properties" in parsed_json:
            properties = parsed_json["properties"]
            properties_list.append(properties)

    coor, gt, coor_system = get_coor_in_space(image_filepath)
    transform_mat = compute_epsg_to_image_mat(coor, gt)

    file = ogr.Open(input_shapefile_filepath)
    assert file is not None, "File {} does not exist!".format(input_shapefile_filepath)
    shape = file.GetLayer(0)
    feature_count = shape.GetFeatureCount()
    polygons = []
    properties_list = []
    for feature_index in tqdm(range(feature_count), desc="Reading features", leave=False):
        feature = shape.GetFeature(feature_index)
        raw_json = feature.ExportToJson()
        parsed_json = json.loads(raw_json)

        # Extract polygon:
        geometry = parsed_json["geometry"]
        if geometry["type"] == "Polygon":
            polygon = np.array(geometry["coordinates"][0])  # TODO: handle polygons with holes (remove [0])
            process_one_polygon(polygon)
        if geometry["type"] == "MultiPolygon":
            for individual_coordinates in geometry["coordinates"]:
                process_one_polygon(np.array(individual_coordinates[0]))  # TODO: handle polygons with holes (remove [0])

    if properties_list:
        return polygons, properties_list
    else:
        return polygons


def create_ogr_polygon(polygon, transform_mat):
    polygon_swapped_coords = polygon_utils.swap_coords(polygon)
    polygon_epsg = apply_transform_mat(polygon_swapped_coords, transform_mat)

    ring = ogr.Geometry(ogr.wkbLinearRing)
    for coord in polygon_epsg:
        ring.AddPoint(coord[0], coord[1])

    # Create polygon
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)
    return poly.ExportToWkt()


def create_ogr_polygons(polygons, transform_mat):
    ogr_polygons = []
    for polygon in polygons:
        ogr_polygons.append(create_ogr_polygon(polygon, transform_mat))
    return ogr_polygons


def save_image_as_geotiff(save_filepath, image, source_geotiff_filepath):
    # Get geo info from source image:
    source_ds = gdal.Open(source_geotiff_filepath)
    if source_ds is None:
        raise FileNotFoundError(f"Could not load source file {source_geotiff_filepath}")
    source_gt = source_ds.GetGeoTransform()
    source_prj = source_ds.GetProjection()

    driver = gdal.GetDriverByName("GTiff")
    outdata = driver.Create(save_filepath, image.shape[1], image.shape[0], image.shape[2])
    outdata.SetGeoTransform(source_gt)  # sets same geotransform as input
    outdata.SetProjection(source_prj)  # sets same projection as input
    for i in range(image.shape[2]):
        outdata.GetRasterBand(i + 1).WriteArray(image[..., i])
    outdata.FlushCache()  # saves to disk!!
    outdata = None  # NOQA
    band = None  # NOQA
    ds = None  # NOQA


def save_shapefile_from_polygons(polygons, image_filepath, output_shapefile_filepath, properties_list=None):
    """
    https://gis.stackexchange.com/a/52708/8104
    """
    assert isinstance(polygons, list) and isinstance(polygons[0], np.ndarray) and \
           len(polygons[0].shape) == 2 and polygons[0].shape[1] == 2, \
            "polygons should be a list of numpy arrays with shape (N, 2)"
    if properties_list is not None:
        assert len(polygons) == len(properties_list), "polygons and properties_list should have the same length"

    coor, gt, coor_system = get_coor_in_space(image_filepath)
    transform_mat = compute_image_to_epsg_mat(coor, gt)
    # Convert polygons to ogr_polygons
    ogr_polygons = create_ogr_polygons(polygons, transform_mat)

    driver = ogr.GetDriverByName('Esri Shapefile')
    ds = driver.CreateDataSource(output_shapefile_filepath)

    # create the spatial reference, WGS84
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(4326)

    layer = ds.CreateLayer('', None, ogr.wkbPolygon)
    # Add one attribute
    field_name_list = []
    field_type_list = []
    if properties_list is not None:
        for properties in properties_list:
            for (key, value) in properties.items():
                if key not in field_name_list:
                    field_name_list.append(key)
                    field_type_list.append(type(value))
    for (name, py_type) in zip(field_name_list, field_type_list):
        if py_type == int:
            ogr_type = ogr.OFTInteger
        elif py_type == float:
            print("is float")
            ogr_type = ogr.OFTReal
        elif py_type == str:
            ogr_type = ogr.OFTString
        else:
            ogr_type = ogr.OFTInteger
        layer.CreateField(ogr.FieldDefn(name, ogr_type))

    defn = layer.GetLayerDefn()

    for index in range(len(ogr_polygons)):
        ogr_polygon = ogr_polygons[index]
        if properties_list is not None:
            properties = properties_list[index]
        else:
            properties = {}

        # Create a new feature (attribute and geometry)
        feat = ogr.Feature(defn)
        for (key, value) in properties.items():
            feat.SetField(key, value)

        # Make a geometry, from Shapely object
        geom = ogr.CreateGeometryFromWkt(ogr_polygon)
        feat.SetGeometry(geom)

        layer.CreateFeature(feat)
        feat = geom = None  # destroy these

    # Save and close everything
    ds = layer = feat = geom = None


def save_shapefile_from_shapely_polygons(polygons, image_filepath, output_shapefile_filepath):
    # Define a polygon feature geometry with one attribute
    schema = {
        'geometry': 'Polygon',
        'properties': {'id': 'int'},
    }
    # shp_crs = "EPSG:4326"
    # shp_srs = Proj(shp_crs)
    raster = rasterio.open(image_filepath)
    # raster_srs = Proj(raster.crs)

    def raster_proj(x, y):
        return raster.transform * (x, y)
    # shp_proj = functools.partial(transform, raster_srs, shp_srs)
    # shp_proj = Transformer.from_proj(raster_srs, shp_srs).transform

    # Write a new Shapefile
    with fiona.open(output_shapefile_filepath, 'w', driver='ESRI Shapefile', schema=schema, crs=fiona.crs.from_epsg(4326)) as c:
        for id, polygon in enumerate(polygons):
            # print("---")
            # print(polygon)
            raster_polygon = shapely.ops.transform(raster_proj, polygon)
            # print(raster_polygon)
            # shp_polygon = shapely.ops.transform(shp_proj, raster_polygon)
            # print(shp_polygon)

            wkt_polygon = shapely.geometry.mapping(raster_polygon)

            c.write({
                'geometry': wkt_polygon,
                'properties': {'id': id},
            })


def indices_of_biggest_intersecting_polygon(polygon_list):
    """
    Assumes polygons which intersect follow each other on the order given by polygon_list.
    This avoids the huge complexity of looking for an intersection between every polygon.

    :param ori_gt_polygons:
    :return:
    """
    keep_index_list = []

    current_cluster = []  # Indices of the polygons belonging to the current cluster (their union has one component)

    for index, polygon in enumerate(polygon_list):
        #  First, check if polygon intersects with current_cluster:
        current_cluster_polygons = [polygon_list[index] for index in current_cluster]
        is_intersection = polygon_utils.check_intersection_with_polygons(polygon, current_cluster_polygons)
        if is_intersection:
            # Just add polygon to the cluster, nothing else to do
            current_cluster.append(index)
        else:
            # This mean the current polygon is part of the next cluster.
            # First, find the biggest polygon in the current cluster
            cluster_max_index = 0
            cluster_max_area = 0
            for cluster_polygon_index in current_cluster:
                cluster_polygon = polygon_list[cluster_polygon_index]
                area = polygon_utils.polygon_area(cluster_polygon)
                if cluster_max_area < area:
                    cluster_max_area = area
                    cluster_max_index = cluster_polygon_index
            # Add index of the biggest polygon to the keep_index_list:
            keep_index_list.append(cluster_max_index)

            # Second, create a new cluster with the current polygon index
            current_cluster = [index]

    return keep_index_list


def get_pixelsize(filepath):
    raster = gdal.Open(filepath)
    gt = raster.GetGeoTransform()
    pixelsize_x = gt[1]
    pixelsize_y = -gt[5]
    pixelsize = (pixelsize_x + pixelsize_y) / 2
    return pixelsize


def crop_shapefile(input_filepath, mask_filepath, output_filepath):
    shp_mask_filepath = os.path.join(os.path.dirname(input_filepath), "mask.shp")
    # ogr2ogr.main(["", "-f", "ESRI Shapefile", shp_mask_filepath, mask_filepath])
    # # ogr2ogr.main(["", "-f", "KML", "-clipsrc", mask_filepath, output_filepath, input_filepath])
    # # script_filepath = os.path.join(os.path.dirname(__file__), "crop_shp_with_shp.sh")
    # # subprocess.Popen(["ogr2ogr", "-clipsrc", mask_filepath, output_filepath, input_filepath])
    #
    # print(input_filepath)
    # print(mask_filepath)
    # print(output_filepath)
    # callstr = ['ogr2ogr',
    #            "-overwrite",
    #            "-t_srs",
    #            "EPSG:27700",
    #            '-clipsrc',
    #            shp_mask_filepath,
    #            output_filepath,
    #            input_filepath,
    #            "-skipfailures"]
    # proc = subprocess.Popen(callstr, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # stdout, stderr = proc.communicate()
    # print(stdout)
    # print(stderr)

    input_file = ogr.Open(input_filepath)
    assert input_file is not None, "File {} does not exist!".format(input_filepath)
    input_layer = input_file.GetLayer(0)
    # for i in range(input_layer.GetFeatureCount()):
    #     feature = input_layer.GetFeature(i)
    #     raw_json = feature.ExportToJson()
    #     parsed_json = json.loads(raw_json)
    #     print(parsed_json)
    #     break

    mask_file = ogr.Open(shp_mask_filepath)
    assert mask_file is not None, "File {} does not exist!".format(shp_mask_filepath)
    mask_layer = mask_file.GetLayer(0)
    print(mask_layer.GetFeatureCount())
    feature = mask_layer.GetFeature(0)
    raw_json = feature.ExportToJson()
    parsed_json = json.loads(raw_json)
    print(parsed_json)

    # create empty result layer
    ogrGeometryType = ogr.Geometry(ogr.wkbPolygon)  # NOQA
    outDriver = ogr.GetDriverByName("ESRI Shapefile")
    outDs = outDriver.CreateDataSource(output_filepath)
    outLayer = outDs.CreateLayer('', None, ogr.wkbPolygon)

    input_layer.Intersection(mask_layer, outLayer, options=["SKIP_FAILURES=YES"])


def main():
    main_dirpath = "/workspace/data/stereo_dataset/raw/leibnitz"
    image_filepath = os.path.join(main_dirpath, "leibnitz_ortho_ref_RGB.tif")
    input_shapefile_filepath = os.path.join(main_dirpath, "Leibnitz_buildings_ref.shp")
    output_shapefile_filepath = os.path.join(main_dirpath, "Leibnitz_buildings_ref.shifted.shp")

    polygons, properties_list = get_polygons_from_shapefile(image_filepath, input_shapefile_filepath)
    print(polygons[0])
    print(properties_list[0])

    # Add shift
    shift = np.array([0, 0])
    shifted_polygons = [polygon + shift for polygon in polygons]
    print(shifted_polygons[0])

    # Save shapefile
    save_shapefile_from_polygons(shifted_polygons, image_filepath, output_shapefile_filepath, properties_list=properties_list)


if __name__ == "__main__":
    main()



def test_predict_override_mean_std():
    """
    Test that the user can override the meanstd at predict time.
    """
    from geowatch.tasks.fusion.predict import predict
    import ubelt as ub
    import os
    from geowatch.utils.lightning_ext.monkeypatches import disable_lightning_hardware_warnings
    disable_lightning_hardware_warnings()
    devices = None
    test_dpath = ub.Path.appdir('geowatch/tests/fusion/').ensuredir()
    results_path = (test_dpath / 'predict').ensuredir()

    # Train a dummy model (TODO: need to build a mechanism so a trained model
    # is created and cached once for all tests that need a model)
    results_path.delete()
    results_path.ensuredir()
    import kwcoco
    train_dset = kwcoco.CocoDataset.demo('special:vidshapes2-gsize64-frames9-speed0.5-multispectral')
    test_dset = kwcoco.CocoDataset.demo('special:vidshapes1-gsize64-frames9-speed0.5-multispectral')
    root_dpath = ub.Path(test_dpath, 'train').ensuredir()
    fit_config = {
        'subcommand': 'fit',
        'fit.data.train_dataset': train_dset.fpath,
        'fit.data.time_steps': 2,
        'fit.data.time_span': "2m",
        'fit.data.chip_dims': 64,
        'fit.data.time_sampling': 'hardish3',
        'fit.data.num_workers': 0,
        #'package_fpath': package_fpath,
        'fit.model.class_path': 'geowatch.tasks.fusion.methods.MultimodalTransformer',
        'fit.model.init_args.global_change_weight': 1.0,
        'fit.model.init_args.global_class_weight': 1.0,
        'fit.model.init_args.global_saliency_weight': 1.0,
        'fit.optimizer.class_path': 'torch.optim.SGD',
        'fit.optimizer.init_args.lr': 1e-5,
        'fit.trainer.max_steps': 10,
        'fit.trainer.accelerator': 'cpu',
        'fit.trainer.devices': 1,
        'fit.trainer.max_epochs': 3,
        'fit.trainer.log_every_n_steps': 1,
        'fit.trainer.default_root_dir': os.fspath(root_dpath),
    }
    from geowatch.tasks.fusion import fit_lightning
    package_fpath = root_dpath / 'final_package.pt'
    fit_lightning.main(fit_config)

    # Unfortunately, its not as easy to get the package path of
    # this call..
    assert ub.Path(package_fpath).exists()

    # Predict via that model
    predict_kwargs = {
        'package_fpath': package_fpath,
        'pred_dataset': ub.Path(results_path) / 'pred.kwcoco.json',
        'test_dataset': test_dset.fpath,
        'datamodule': 'KWCocoVideoDataModule',
        'batch_size': 1,
        'num_workers': 0,
        'devices': devices,
        'override_meanstd': ub.codeblock(
            '''
            - sensor: '*'
              channels: 'B1|B10|B11|B8|B8a'
              mean: [0, 0, 0, 0, 0]
              std: [1, 1, 1, 1, 1]
            '''),
        'draw_batches': 1,
    }
    result_dataset = predict(**predict_kwargs)
    dset = result_dataset
    dset.dataset['info'][-1]['properties']['config']['time_sampling']
    # Check that the result format looks correct
    for vidid in dset.index.videos.keys():
        # Note: only some of the images in the pred sequence will get
        # a change predictoion, depending on the temporal sampling.
        images = dset.images(dset.index.vidid_to_gids[1])
        pred_chans = [[a['channels'] for a in aux] for aux in images.lookup('auxiliary')]
        assert any('change' in cs for cs in pred_chans), 'some frames should have change'
        assert not all('change' in cs for cs in pred_chans), 'some frames should not have change'
        # Test number of annots in each frame
        frame_to_cathist = {
            img['frame_index']: ub.dict_hist(annots.cnames, labels=result_dataset.object_categories())
            for img, annots in zip(images.objs, images.annots)
        }
        assert frame_to_cathist[0]['change'] == 0, 'first frame should have no change polygons'
        # This test may fail with very low probability, so warn
        import warnings
        if sum(d['change'] for d in frame_to_cathist.values()) == 0:
            warnings.warn('should have some change predictions elsewhere')
    coco_img = dset.images().coco_images[1]
    # Test that new quantization does not existing APIs
    pred1 = coco_img.imdelay('salient', nodata_method='float').finalize()
    assert pred1.max() <= 1
    # new delayed image does not make it easy to remove dequantization
    # add test back in if we add support for that.



import numpy as np
import time
import sklearn.datasets
import skimage.transform
import scipy.stats

from . import python_utils
from . import image_utils

# if python_utils.module_exists("matplotlib.pyplot"):
#     import matplotlib.pyplot as plt

CV2 = False
if python_utils.module_exists("cv2"):
    import cv2

    CV2 = True


# import multiprocessing
#
# import python_utils
#
# if python_utils.module_exists("joblib"):
#     from joblib import Parallel, delayed
#     JOBLIB = True
# else:
#     JOBLIB = False


# def plot_field_map(field_map):
#     from mpl_toolkits.mplot3d import Axes3D
#
#     row = np.linspace(0, 1, field_map.shape[0])
#     col = np.linspace(0, 1, field_map.shape[1])
#     rr, cc = np.meshgrid(row, col, indexing='ij')
#
#     fig = plt.figure(figsize=(18, 9))
#     ax = fig.add_subplot(121, projection='3d')
#     ax.plot_surface(rr, cc, field_map[:, :, 0], rstride=3, cstride=3, linewidth=1, antialiased=True)
#
#     ax = fig.add_subplot(122, projection='3d')
#     ax.plot_surface(rr, cc, field_map[:, :, 1], rstride=3, cstride=3, linewidth=1, antialiased=True)
#
#     plt.show()

# --- Classes --- #

class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, name="", init_val=0, fmt=':f'):
        self.name = name
        self.init_val = init_val
        self.fmt = fmt
        self.val = self.avg = self.init_val
        self.sum = self.count = 0

    def reset(self):
        self.val = self.avg = self.init_val
        self.sum = self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def get_avg(self):
        return self.avg

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)


class RunningDecayingAverage(object):
    """
    Updates average with val*(1 - decay) + avg*decay
    """

    def __init__(self, decay, init_val=0):
        assert 0 < decay < 1
        self.decay = decay
        self.init_val = init_val
        self.val = self.avg = self.init_val

    def reset(self):
        self.val = self.avg = self.init_val

    def update(self, val):
        self.val = val
        self.avg = (1 - self.decay) * val + self.decay * self.avg

    def get_avg(self):
        return self.avg


class DispFieldMapsPatchCreator:
    def __init__(self, global_shape, patch_res, map_count,
                 modes, gauss_mu_range, gauss_sig_scaling):
        self.global_shape = global_shape
        self.patch_res = patch_res
        self.map_count = map_count
        self.modes = modes
        self.gauss_mu_range = gauss_mu_range
        self.gauss_sig_scaling = gauss_sig_scaling

        self.current_patch_index = -1
        self.patch_boundingboxes = image_utils.compute_patch_boundingboxes(self.global_shape, stride=self.patch_res,
                                                                           patch_res=self.patch_res)
        self.disp_maps = None
        self.create_new_disp_maps()

    def create_new_disp_maps(self):
        print("DispFieldMapsPatchCreator.create_new_disp_maps()")
        self.disp_maps = create_displacement_field_maps(self.global_shape, self.map_count, self.modes,
                                                        self.gauss_mu_range, self.gauss_sig_scaling)

    def get_patch(self):
        self.current_patch_index += 1

        if len(self.patch_boundingboxes) <= self.current_patch_index:
            self.current_patch_index = 0
            self.create_new_disp_maps()

        patch_boundingbox = self.patch_boundingboxes[self.current_patch_index]
        patch_disp_maps = self.disp_maps[:, patch_boundingbox[0]:patch_boundingbox[2],
                                         patch_boundingbox[1]:patch_boundingbox[3], :]
        return patch_disp_maps


# --- --- #

def compute_crossfield_c0c2(u, v):
    c0 = np.power(u, 2) * np.power(v, 2)
    c2 = - (np.power(u, 2) + np.power(v, 2))
    crossfield = np.stack([c0.real, c0.imag, c2.real, c2.imag], axis=-1)
    return crossfield


def compute_crossfield_uv(c0c2):
    c0 = c0c2[..., 0] + 1j * c0c2[..., 1]
    c2 = c0c2[..., 2] + 1j * c0c2[..., 3]
    sqrt_c2_squared_minus_4c0 = np.sqrt(np.power(c2, 2) - 4 * c0)
    u_squared = (c2 + sqrt_c2_squared_minus_4c0) / 2
    v_squared = (c2 - sqrt_c2_squared_minus_4c0) / 2
    u = np.sqrt(u_squared)
    v = np.sqrt(v_squared)
    return u, v


def to_homogeneous(array):
    new_array = np.ones(
        (array.shape[0],
         array.shape[1] + 1),
        dtype=array.dtype)
    new_array[..., :-1] = array
    return new_array


def to_euclidian(array_homogeneous):
    array = array_homogeneous[:, 0:2] / array_homogeneous[:, 2:3]
    return array


def stretch(array):
    mini = np.min(array)
    maxi = np.max(array)
    if maxi - mini:
        array -= mini
        array *= 2 / (maxi - mini)
        array -= 1
    return array


def crop_center(array, out_shape):
    assert len(out_shape) == 2, "out_shape should be of length 2"
    in_shape = np.array(array.shape[:2])
    start = in_shape // 2 - (out_shape // 2)
    out_array = array[start[0]:start[0] + out_shape[0],
                      start[1]:start[1] + out_shape[1], ...]
    return out_array


def multivariate_gaussian(pos, mu, sigma):
    """Return the multivariate Gaussian distribution on array pos.

    pos is an array constructed by packing the meshed arrays of variables
    x_1, x_2, x_3, ..., x_k into its _last_ dimension.

    """

    n = mu.shape[0]
    sigma_det = np.linalg.det(sigma)
    sigma_inv = np.linalg.inv(sigma)
    N = np.sqrt((2 * np.pi) ** n * sigma_det)
    # This einsum call calculates (x-mu)T.sigma-1.(x-mu) in a vectorized
    # way across all the input variables.

    # print("\tStarting to create multivariate Gaussian")
    # start = time.time()

    # print((pos - mu).shape)
    # print(sigma_inv.shape)
    try:
        fac = np.einsum(
            '...k,kl,...l->...',
            pos - mu,
            sigma_inv,
            pos - mu,
            optimize=True)
    except BaseException:
        fac = np.einsum('...k,kl,...l->...', pos - mu, sigma_inv, pos - mu)
    # print(fac.shape)

    # end = time.time()
    # print("\tFinished Gaussian in {}s".format(end - start))

    return np.exp(-fac / 2) / N


def create_multivariate_gaussian_mixture_map(
        shape, mode_count, mu_range, sig_scaling):
    shape = np.array(shape)
    # print("Starting to create multivariate Gaussian mixture")
    # main_start = time.time()

    dim_count = 2
    downsample_factor = 4
    dtype = np.float32

    mu_scale = mu_range[1] - mu_range[0]
    row = np.linspace(
        mu_range[0],
        mu_range[1],
        mu_scale *
        shape[0] /
        downsample_factor,
        dtype=dtype)
    col = np.linspace(
        mu_range[0],
        mu_range[1],
        mu_scale *
        shape[1] /
        downsample_factor,
        dtype=dtype)
    rr, cc = np.meshgrid(row, col, indexing='ij')
    grid = np.stack([rr, cc], axis=2)

    mus = np.random.uniform(
        mu_range[0],
        mu_range[1],
        (mode_count,
         dim_count,
         2)).astype(dtype)
    # gams = np.random.rand(mode_count, dim_count, 2, 2).astype(dtype)
    signs = np.random.choice([1, -1], size=(mode_count, dim_count))

    # print("\tAdding gaussian mixtures one by one")
    # start = time.time()

    # if JOBLIB:
    #     # Parallel computing of multivariate gaussians
    #     inputs = range(8)
    #
    #     def processInput(i):
    #         size = 10 * i + 2000
    #         a = np.random.random_sample((size, size))
    #         b = np.random.random_sample((size, size))
    #         n = np.dot(a, b)
    #         return n
    #
    #     num_cores = multiprocessing.cpu_count()
    #     print("num_cores: {}".format(num_cores))
    #     # num_cores = 1
    #
    #     results = Parallel(n_jobs=num_cores)(delayed(processInput)(i) for i in inputs)
    #     for result in results:
    #         print(result.shape)
    #
    #     gaussian_mixture = np.zeros_like(grid)
    # else:
    gaussian_mixture = np.zeros_like(grid)
    for mode_index in range(mode_count):
        for dim in range(dim_count):
            sig = (sig_scaling[1] - sig_scaling[0]) * \
                sklearn.datasets.make_spd_matrix(2) + sig_scaling[0]
            # sig = (sig_scaling[1] - sig_scaling[0]) * np.dot(gams[mode_index, dim], np.transpose(gams[mode_index, dim])) + sig_scaling[0]
            sig = sig.astype(dtype)
            multivariate_gaussian_grid = signs[mode_index, dim] * \
                multivariate_gaussian(grid, mus[mode_index, dim], sig)
            gaussian_mixture[:, :, dim] += multivariate_gaussian_grid

    # end = time.time()
    # print("\tFinished adding gaussian mixtures in {}s".format(end - start))

    # squared_gaussian_mixture = np.square(gaussian_mixture)
    # magnitude_disp_field_map = np.sqrt(squared_gaussian_mixture[:, :, 0] + squared_gaussian_mixture[:, :, 1])
    # max_magnitude = magnitude_disp_field_map.max()

    gaussian_mixture[:, :, 0] = stretch(gaussian_mixture[:, :, 0])
    gaussian_mixture[:, :, 1] = stretch(gaussian_mixture[:, :, 1])

    # Crop
    gaussian_mixture = crop_center(
        gaussian_mixture,
        shape // downsample_factor)

    # plot_field_map(gaussian_mixture)

    # Upsample mixture
    # gaussian_mixture = skimage.transform.rescale(gaussian_mixture, downsample_factor)
    gaussian_mixture = skimage.transform.resize(gaussian_mixture, shape)

    main_end = time.time()  # NOQA
    # print("Finished multivariate Gaussian mixture in {}s".format(main_end - main_start))

    return gaussian_mixture


def create_displacement_field_maps(
        shape, map_count, modes, gauss_mu_range, gauss_sig_scaling, seed=None):
    if seed is not None:
        np.random.seed(seed)
    disp_field_maps_list = []
    for disp_field_map_index in range(map_count):
        disp_field_map_normed = create_multivariate_gaussian_mixture_map(shape,
                                                                         modes,
                                                                         gauss_mu_range,
                                                                         gauss_sig_scaling)
        disp_field_maps_list.append(disp_field_map_normed)
    disp_field_maps = np.stack(disp_field_maps_list, axis=0)

    return disp_field_maps


def get_h_mat(t, theta, scale_offset, shear, p):
    """
    Computes the homography matrix given the parameters
    See https://medium.com/uruvideo/dataset-augmentation-with-random-homographies-a8f4b44830d4
    (fixed mistake in H_a)

    :param t: 2D translation vector
    :param theta: Scalar angle
    :param scale_offset: 2D scaling vector
    :param shear: 2D shearing vector
    :param p: 2D projection vector
    :return: h_mat: shape (3, 3)
    """
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)
    h_e = np.array([
        [cos_theta, -sin_theta, t[0]],
        [sin_theta, cos_theta, t[1]],
        [0, 0, 1],
    ])
    h_a = np.array([
        [1 + scale_offset[0], shear[1], 0],
        [shear[0], 1 + scale_offset[1], 0],
        [0, 0, 1],
    ])
    h_p = np.array([
        [1, 0, 0],
        [0, 1, 0],
        [p[0], p[1], 1],
    ])
    h_mat = h_e @ h_a @ h_p
    return h_mat


if CV2:
    def find_homography_4pt(src, dst):
        """
        Estimates the homography that transforms src points into dst points.
        Then converts the matrix representation into the 4 points representation.

        :param src:
        :param dst:
        :return:
        """
        h_mat, _ = cv2.findHomography(src, dst)
        h_4pt = convert_h_mat_to_4pt(h_mat)
        return h_4pt

    def convert_h_mat_to_4pt(h_mat):
        src_4pt = np.array([[
            [-1, -1],
            [1, -1],
            [1, 1],
            [-1, 1],
        ]], dtype=np.float64)
        h_4pt = cv2.perspectiveTransform(src_4pt, h_mat)
        return h_4pt

    def convert_h_4pt_to_mat(h_4pt):
        src_4pt = np.array([
            [-1, -1],
            [1, -1],
            [1, 1],
            [-1, 1],
        ], dtype=np.float32)
        h_4pt = h_4pt.astype(np.float32)
        h_mat = cv2.getPerspectiveTransform(src_4pt, h_4pt)
        return h_mat

    def field_map_to_image(field_map):
        mag, ang = cv2.cartToPolar(field_map[..., 0], field_map[..., 1])
        hsv = np.zeros((field_map.shape[0], field_map.shape[1], 3))
        hsv[..., 0] = ang * 180 / np.pi / 2
        hsv[..., 1] = 255
        hsv[..., 2] = cv2.normalize(mag, None, 0, 255, cv2.NORM_MINMAX)
        hsv = hsv.astype(np.uint8)
        rgb = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)
        return rgb
else:
    def find_homography_4pt(src, dst):
        print(
            "cv2 is not available, the find_homography_4pt(src, dst) function cannot work!")

    def convert_h_mat_to_4pt(h_mat):
        print("cv2 is not available, the convert_h_mat_to_4pt(h_mat) function cannot work!")

    def convert_h_4pt_to_mat(h_4pt):
        print("cv2 is not available, the convert_h_4pt_to_mat(h_4pt) function cannot work!")

    def field_map_to_image(field_map):
        print(
            "cv2 is not available, the field_map_to_image(field_map) function cannot work!")


def circular_diff(a1, a2, range_max):
    """
    Compute difference between a1 and a2 belonging to the circular interval [0, range_max).
    For example to compute angle difference, use range_max=2*PI.
    a1 and a2 must be between range_min and range_max!
    Thus difference between 0 and range_max is 0.
    :param a1: numpy array
    :param a2: numpy array
    :param range_max:
    :return:
    """
    d = range_max / 2 - np.abs(np.abs(a1 - a2) - range_max / 2)
    return d


def invert_permutation(p):
    '''The argument p is assumed to be some permutation of 0, 1, ..., len(p)-1.
    Returns an array s, where s[i] gives the index of i in p.
    '''
    s = np.empty(p.size, p.dtype)
    s[p] = np.arange(p.size)
    return s


def region_growing_1d(array, max_range, max_skew):
    """
    :param array:
    :param max_var:
    :param max_mean_median_diff:
    :return:
    """

    def verify_predicate(region):
        """
        Region is sorted
        :param region:
        :return:
        """
        skew = scipy.stats.skew(region)
        return region[-1] - region[0] < max_range and abs(skew) < max_skew

    assert len(
        array.shape) == 1, "array should be 1d, not {}".format(
        array.shape)
    p = np.argsort(array)
    sorted_array = array[p]

    labels = np.zeros(len(sorted_array), dtype=np.long)
    region_start = 0
    region_label = 1
    labels[region_start] = region_label
    centers = []
    for i in range(1, len(sorted_array)):
        region = sorted_array[region_start:i + 1]
        if not verify_predicate(region):
            # End current region
            median = region[len(region) // 2]  # region is sorted
            centers.append(median)
            # Begin a new region
            region_start = i
            region_label += 1
        labels[i] = region_label
    centers.append(median)

    return labels[invert_permutation(p)], centers


def bilinear_interpolate(im, pos):
    # From
    # https://gist.github.com/peteflorence/a1da2c759ca1ac2b74af9a83f69ce20e
    x = pos[..., 1]
    y = pos[..., 0]

    x0 = np.floor(x).astype(int)
    x1 = x0 + 1
    y0 = np.floor(y).astype(int)
    y1 = y0 + 1

    x0_clipped = np.clip(x0, 0, im.shape[1] - 1)
    x1_clipped = np.clip(x1, 0, im.shape[1] - 1)
    y0_clipped = np.clip(y0, 0, im.shape[0] - 1)
    y1_clipped = np.clip(y1, 0, im.shape[0] - 1)

    Ia = im[y0_clipped, x0_clipped]
    Ib = im[y1_clipped, x0_clipped]
    Ic = im[y0_clipped, x1_clipped]
    Id = im[y1_clipped, x1_clipped]

    wa = (x1 - x) * (y1 - y)
    wb = (x1 - x) * (y - y0)
    wc = (x - x0) * (y1 - y)
    wd = (x - x0) * (y - y0)

    value = (Ia.T * wa).T + (Ib.T * wb).T + (Ic.T * wc).T + (Id.T * wd).T

    return value


def main():
    import matplotlib.pyplot as plt
    # shape = (220, 220)
    # mode_count = 30
    # mu_range = [0, 1]
    # sig_scaling = [0.0, 0.002]
    # create_multivariate_gaussian_mixture_map(shape, mode_count, mu_range, sig_scaling)

    # a1 = np.array([0.0])
    # a2 = np.array([3*np.pi/4])
    # range_max = np.pi
    # d = circular_diff(a1, a2, range_max)
    # print(d)

    array = np.concatenate(
        [np.arange(1, 1.01, 0.001), np.arange(0, np.pi / 2, np.pi / 100)])
    print(array)
    labels = region_growing_1d(array, max_range=np.pi / 10, max_skew=1)
    print(labels)

    plt.plot(array, labels, ".")
    plt.show()


if __name__ == "__main__":
    main()





def test_tracker_time_split_thresh():
    """
    CommandLine:
        xdoctest tests/test_recurring_sites_in_tracker.py test_tracker_time_split_thresh
    """
    import json
    import os
    import kwcoco
    import ubelt as ub
    from geowatch.cli import run_tracker

    coco_dset = build_recurring_sites_coco()

    dpath = ub.Path.appdir('geowatch', 'test', 'tracking', 'recuring_sites').ensuredir()
    dpath.delete().ensuredir()

    regions_dir = dpath / 'regions/'
    bas_coco_fpath = dpath / 'bas_output.kwcoco.json'
    bas_fpath = dpath / 'bas_sites.json'

    viz_outdir = str((dpath / 'track_viz').ensuredir())

    # Run BAS
    bas_argv = [
        '--in_file', os.fspath(coco_dset.fpath),
        '--out_site_summaries_dir', str(regions_dir),
        '--out_site_summaries_fpath',  str(bas_fpath),
        '--out_kwcoco', str(bas_coco_fpath),
        '--track_fn', 'saliency_heatmaps',
        '--sensor_warnings', '0',
        '--viz_out_dir', os.fspath(viz_outdir),
        '--track_kwargs', json.dumps({
            'thresh': 0.3,
            'time_thresh': .8,
            'min_area_square_meters': None,
            'moving_window_size': 2,
            'max_area_square_meters': None,
            'polygon_simplify_tolerance': 1,
            'time_split_thresh': 0.5,
        }),
    ]
    run_tracker.main(argv=bas_argv)

    bas_coco_dset = kwcoco.CocoDataset(bas_coco_fpath)

    ann_trackids = bas_coco_dset.annots().lookup('track_id', None)
    ann_timestamps = bas_coco_dset.annots().images.lookup('date_captured', None)
    track_times = ub.group_items(ann_timestamps, ann_trackids)
    print('track_times = {}'.format(ub.urepr(track_times, nl=1)))
    assert len(track_times) == 3

    if 0:
        ub.cmd(f'geowatch visualize {bas_coco_dset.fpath} --stack=only --smart', system=1)


def build_recurring_sites_coco():
    """
    Build a simple test case where the site blips in and out every year.

    CommandLine:
        xdoctest tests/test_recurring_sites_in_tracker.py build_recurring_sites_coco
    """
    import geowatch
    import ubelt as ub
    import kwarray
    import kwimage
    import numpy as np
    from kwutil import util_time
    dpath = ub.Path.appdir('geowatch/tests/recurring_sites')
    geowatch.coerce_kwcoco('geowatch-msi')
    asset_dpath = (dpath / 'assets').ensuredir()

    rng = kwarray.ensure_rng(0)

    gh, gw = 512, 512
    num_frames = 50

    poly = kwimage.Polygon.random(rng=rng).scale((gw, gh))
    poly = poly.scale(0.3, about='centroid')

    start_time = util_time.coerce_datetime('2020-01-01')
    frame_rate = util_time.coerce_timedelta('1week')

    current_time = start_time

    blip_times = [4, 5, 6, 16, 17, 18, 40, 41, 42]

    from geowatch.demo.smart_kwcoco_demodata import _random_utm_box
    utm_box, utm_crs_info = _random_utm_box()
    auth = utm_crs_info['auth']
    assert auth[0] == 'EPSG'
    epsg_int = int(auth[1])
    ulx, uly, lrx, lry = utm_box.to_ltrb().data[0]

    import kwcoco
    coco_dset = kwcoco.CocoDataset()
    video_id = coco_dset.add_video('video1', width=gw, height=gh)

    for frame_idx in range(num_frames):
        # Write image to disk
        name = f'video_{video_id:03d}_frame_{frame_idx:03d}'
        fname = f'{name}.tif'
        gpath = asset_dpath / fname
        # canvas = np.zeros((gh, gw, 1), dtype=np.float32)
        canvas = rng.rand(gh, gw, 1).astype(np.float32) * 0.1
        if frame_idx in blip_times:
            # Draw blips only sometimes
            canvas = poly.fill(canvas, value=1)
        kwimage.imwrite(gpath, canvas, backend='gdal')
        command = f'gdal_edit.py -a_ullr {ulx} {uly} {lrx} {lry} -a_srs EPSG:{epsg_int} {gpath}'
        ub.cmd(command, shell=True, check=True)

        # Add base container for the image
        image = {
            'video_id': video_id,
            'name': name,
            'file_name': None,
            'sensor_coarse': 'S2',
            'date_captured': current_time.date().isoformat(),
            'frame_index': frame_idx,
            'width': gw,
            'height': gh,
        }
        gid = coco_dset.add_image(**image)
        # Use the coco image to add the asset
        coco_img = coco_dset.coco_image(gid)
        coco_img.add_asset(
            file_name=gpath,
            channels='salient',
            width=gw,
            height=gh
        )
        current_time += frame_rate

    from geowatch.utils import kwcoco_extensions
    # Do a consistent transfer of the hacked seeded geodata to the other images
    kwcoco_extensions.ensure_transfered_geo_data(coco_dset)
    kwcoco_extensions.coco_populate_geo_heuristics(coco_dset)
    kwcoco_extensions.warp_annot_segmentations_to_geos(coco_dset)
    kwcoco_extensions.populate_watch_fields(
        coco_dset, target_gsd=0.3, enable_valid_region=True,
        overwrite=True)

    coco_dset.fpath = dpath / 'test.kwcoco.zip'
    coco_dset.dump()

    if 0:
        ub.cmd(f'geowatch visualize {coco_dset.fpath} --stack=only --smart', system=1)

    return coco_dset



r"""
Transfering cold features

CommandLine:

    DATA_DVC_DPATH=$(geowatch_dvc --tags=phase2_data --hardware="auto")
    EXPT_DVC_DPATH=$(geowatch_dvc --tags=phase2_expt --hardware="auto")
    python -m geowatch.tasks.cold.transfer_features \
        --coco_fpath="$DATA_DVC_DPATH/Drop6/imgonly_KR_R001_cold-HTR.kwcoco.zip" \
        --combine_fpath="$DATA_DVC_DPATH/Drop6-MeanYear10GSD-V2/imgonly-KR_R001.kwcoco.zip" \
        --new_coco_fpath="$DATA_DVC_DPATH/Drop6-MeanYear10GSD-V2/imganns-KR_R001_uconn_cold.kwcoco.zip"

"""
import os
import ubelt as ub
import kwcoco
import kwimage
from os.path import join
from kwutil import util_time
import scriptconfig as scfg

try:
    from line_profiler import profile
except ImportError:
    profile = ub.identity


class TransferCocoConfig(scfg.DataConfig):
    """
    Transfer channels to one kwcoco file to the nearest image in the future.
    """

    src_kwcoco = scfg.Value(None, position=1, help=ub.paragraph(
        '''
        a path to a file to input kwcoco file (to predict on)
        '''), alias=['coco_fpath'])
    dst_kwcoco = scfg.Value(None, help=ub.paragraph(
        '''
        a path to a file to combined input kwcoco file (to merge with)
        '''), alias=['combine_fpath'])
    new_coco_fpath = scfg.Value(None, help='file path for modified output coco json')
    channels_to_transfer = scfg.Value(None, help='COLD channels for transfer')

    io_workers = scfg.Value(0, help='number of workers for copy-asset jobs')

    copy_assets = scfg.Value(False, help='if True copy the assests to the new bundle directory')

    respect_sensors = scfg.Value(
        True, help='if True only transfer features to images that share a sensor')

    # TODO: propogate strategy?
    max_propogate = scfg.Value(1, help='maximum number of future images a src image can transfer onto.')

    allow_affine_approx = scfg.Value(False, isflag=True, help=ub.paragraph(
        '''
        if True allow a transfer between different CRSs by approximating an affine transform. This should only be used for quick-and-dirty analysis, and never in production
        '''))


@profile
def transfer_features_main(cmdline=1, **kwargs):
    """
    Args:
        cmdline (int, optional): _description_. Defaults to 1.

    Ignore:
        python -m geowatch.tasks.cold.transfer_features --help
        TEST_COLD=1 xdoctest -m geowatch.tasks.cold.transfer_features transfer_features_main

     Example:
        >>> # xdoctest: +REQUIRES(env:TEST_COLD)
        >>> from geowatch.tasks.cold.transfer_features import transfer_features_main
        >>> from geowatch.tasks.cold.transfer_features import *
        >>> kwargs= dict(
        >>>   coco_fpath = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6/imgonly_KR_R001_cold-V2.kwcoco.zip'),
        >>>   combine_fpath = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6-MeanYear10GSD-V2/imgonly-KR_R001.kwcoco.zip'),
        >>>   new_coco_fpath = ub.Path('/gpfs/scratchfs1/zhz18039/jws18003/new-repos/smart_data_dvc2/Drop6-MeanYear10GSD-V2/imganns-KR_R001_uconn_cold_test.kwcoco.zip'),
        >>>   #workermode = 'process',
        >>> )
        >>> cmdline=0
        >>> transfer_features_main(cmdline, **kwargs)

    Example:
        >>> from geowatch.tasks.cold.transfer_features import transfer_features_main
        >>> from geowatch.tasks.cold.transfer_features import *
        >>> import geowatch
        >>> dset1 = geowatch.coerce_kwcoco('geowatch-msi', geodata=True, heatmap=True, dates=True)
        >>> dset2 = dset1.copy()
        >>> # Remove saliency assets from dset2
        >>> for img in dset2.images().coco_images:
        >>>     asset = img.find_asset_obj('salient')
        >>>     img['auxiliary'].remove(asset)
        >>> dset2_orig = dset2.copy()
        >>> # Transfer the saliency from dset1 onto dset2
        >>> kwargs = TransferCocoConfig(**{
        >>>     'src_kwcoco': dset1,
        >>>     'dst_kwcoco': dset2,
        >>>     'new_coco_fpath': 'return',
        >>>     'channels_to_transfer': 'salient',
        >>>     'copy_assets': False,
        >>> })
        >>> cmdline = False
        >>> new_dset = transfer_features_main(cmdline, **kwargs)
        >>> assert new_dset is dset2, 'modifies combine_fpath inplace'
        >>> from geowatch.utils import kwcoco_extensions
        >>> stats1 = kwcoco_extensions.coco_channel_stats(dset1)
        >>> stats2 = kwcoco_extensions.coco_channel_stats(dset2)
        >>> stats2_orig = kwcoco_extensions.coco_channel_stats(dset2_orig)
        >>> assert stats2['single_chan_hist']['salient'] == stats1['single_chan_hist']['salient'], 'all assets should transfer'
        >>> assert stats2['single_chan_hist']['salient'] == dset2.n_images, 'all dst images should get saliency here'
        >>> assert stats2_orig['single_chan_hist']['salient'] == 0
    """
    #NOTE: This script doesn't consider timestamp = True
    config = TransferCocoConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich
    rich.print('config = {}'.format(ub.urepr(config, nl=1)))

    from geowatch.cli.reproject_annotations import keyframe_interpolate
    from geowatch.utils import process_context
    from kwutil import util_json

    resolved_config = config.to_dict()
    resolved_config = util_json.ensure_json_serializable(resolved_config)
    proc_context = process_context.ProcessContext(
        name='geowatch.tasks.cold.transfer_features',
        type='process',
        config=resolved_config,
    )
    proc_context.start()

    # Assign variables
    default_channels = [
        'blue_COLD_cv', 'green_COLD_cv', 'red_COLD_cv', 'nir_COLD_cv',
        'swir16_COLD_cv', 'swir22_COLD_cv', 'blue_COLD_a0', 'green_COLD_a0',
        'red_COLD_a0', 'nir_COLD_a0', 'swir16_COLD_a0', 'swir22_COLD_a0',
        'blue_COLD_a1', 'green_COLD_a1', 'red_COLD_a1', 'nir_COLD_a1',
        'swir16_COLD_a1', 'swir22_COLD_a1', 'blue_COLD_b1', 'green_COLD_b1',
        'red_COLD_b1', 'nir_COLD_b1', 'swir16_COLD_b1', 'swir22_COLD_b1',
        'blue_COLD_c1', 'green_COLD_c1', 'red_COLD_c1', 'nir_COLD_c1',
        'swir16_COLD_c1', 'swir22_COLD_c1', 'blue_COLD_rmse',
        'green_COLD_rmse', 'red_COLD_rmse', 'nir_COLD_rmse',
        'swir16_COLD_rmse', 'swir22_COLD_rmse'
    ]

    do_return = (config['new_coco_fpath'] == 'return')
    if not do_return:
        new_coco_fpath = ub.Path(config['new_coco_fpath'])

    if config['channels_to_transfer'] is None:
        channels_to_transfer = kwcoco.FusedChannelSpec.coerce(default_channels)
    else:
        channels_to_transfer = kwcoco.FusedChannelSpec.coerce(config['channels_to_transfer'])

    print('Loading source kwcoco file')
    src = kwcoco.CocoDataset.coerce(config['src_kwcoco'])

    print('Loading destination kwcoco file')
    dst = kwcoco.CocoDataset.coerce(config['dst_kwcoco'])

    print(f'src={src}')
    print(f'dst={dst}')

    if not do_return:
        _update_coco_fpath(dst, new_coco_fpath)

    src_video_names = src.videos().lookup('name')

    dst_vidnames = sorted(set(dst.index.name_to_video))
    src_vidnames = sorted(set(src.index.name_to_video))
    if src_vidnames != dst_vidnames:
        # If video names do not agree, we need to check for overlaps
        from geowatch.utils import kwcoco_extensions
        from kwgis.utils import util_gis
        src_vid_gdf = kwcoco_extensions.covered_video_geo_regions(src)
        dst_vid_gdf = kwcoco_extensions.covered_video_geo_regions(dst)
        dst_to_src_idxs = util_gis.geopandas_pairwise_overlaps(dst_vid_gdf, src_vid_gdf)
        # For each site, chose a single video assign to Note: this only works
        # well when the dst are smaller than the src, which is the case it is
        # being written for. If larger dst are needed the logic will need to
        # change.
        vidname_pairs = []
        for dst_idx, src_idxs in dst_to_src_idxs.items():
            if len(src_idxs) == 0:
                continue
            elif len(src_idxs) == 1:
                idx = 0
            else:
                qshape = dst_vid_gdf.iloc[dst_idx]['geometry']
                candidates = src_vid_gdf.iloc[src_idxs]
                overlaps = []
                for dshape in candidates['geometry']:
                    iarea = qshape.intersection(dshape).area
                    uarea = qshape.area
                    iofa = iarea / uarea
                    overlaps.append(iofa)
                idx = ub.argmax(overlaps)

            dst_vidname = dst_vid_gdf.iloc[dst_idx].video_name
            src_vidname = src_vid_gdf.iloc[src_idxs[idx]].video_name
            vidname_pairs.append((src_vidname, dst_vidname))
    else:
        vidname_pairs = list(zip(src_video_names, dst_vidnames))
        ...

    # We will build a list containing all of the assignments to make
    assignments = []
    for src_vidname, dst_vidname in vidname_pairs:
        # For each corresponding video
        src_video = src.index.name_to_video[src_vidname]
        dst_video = dst.index.name_to_video[dst_vidname]
        # Look at each sequence of images
        all_src_images = src.images(video_id=src_video['id'])
        dst_images = dst.images(video_id=dst_video['id'])

        # Filter out the source images missing the channels we want to transfer
        keep_flags = [
            coco_img.channels.intersection(channels_to_transfer).numel() > 0
            for coco_img in all_src_images.coco_images
        ]
        src_images = all_src_images.compress(keep_flags)

        # Build a list of columns used as group-ids which will be used to
        # ensure src images only transfer to dst images with the same group-id.
        src_groupers = []
        dst_groupers = []

        if config.respect_sensors:
            # Use sensors in group-ids, so we only transfer between similar
            # sensors
            src_sensors = src_images.lookup('sensor_coarse')
            dst_sensors = dst_images.lookup('sensor_coarse')
            src_groupers.append(src_sensors)
            dst_groupers.append(dst_sensors)

        if src_groupers:
            src_groupids = zip(*src_groupers)
            dst_groupids = zip(*dst_groupers)
            groupid_to_src_gids = ub.group_items(src_images, src_groupids)
            groupid_to_dst_gids = ub.group_items(dst_images, dst_groupids)
        else:
            # No groupers were given, so any src can transfer to any dst.
            groupid_to_src_gids = {'__nogroup__': src_images}
            groupid_to_dst_gids = {'__nogroup__': dst_images}

        for groupid in groupid_to_src_gids.keys():
            src_image_ids = groupid_to_src_gids[groupid]
            dst_image_ids = groupid_to_dst_gids[groupid]

            src_subimgs = src.images(src_image_ids)
            dst_subimgs = dst.images(dst_image_ids)

            # Find the timestamps for each image in both sequences
            src_timestamps = [util_time.coerce_datetime(d) for d in src_subimgs.lookup('date_captured')]
            dst_timestamps = [util_time.coerce_datetime(d) for d in dst_subimgs.lookup('date_captured')]
            assert sorted(src_timestamps) == src_timestamps, 'data should be in order'
            assert sorted(dst_timestamps) == dst_timestamps, 'data should be in order'

            # We will use logic in reproject annotations to assign images with COLD
            # features to the nearest image in the future. In this case the
            # keyframes are the images with COLD features and the target sequence
            # are the images we are transfering onto.
            target_times = dst_timestamps

            key_infos = [{'time': d, 'applies': 'future'} for d in src_timestamps]
            assigned_indexes = keyframe_interpolate(target_times, key_infos)
            # The above function returns a list for each key-frame with the
            # assigned indexes but we are only going to assign it to one of them,
            # so choose the earliest one in each group. The minimum index will be
            # the earliest because dates are sorted.
            for src_idx, dst_idxs in enumerate(assigned_indexes):
                if dst_idxs:
                    dst_idxs = sorted(dst_idxs)
                    chosen_dst_idxs = dst_idxs[:config.max_propogate]
                    for dst_idx in chosen_dst_idxs:
                        assignments.append({
                            'src_image_id': src_image_ids[src_idx],
                            'dst_image_id': dst_image_ids[dst_idx],
                            'src_vidname': src_vidname,
                            'dst_vidname': dst_vidname,
                        })

    print(f'Found {len(assignments)} image-to-image assignments to transfer')
    assets_to_transfer = []

    # Now we have image assignments, find the assets to transfer
    vidname_to_assignments = ub.group_items(assignments, lambda x: (x['src_vidname'], x['dst_vidname']))
    for vidnames, vid_assignments in vidname_to_assignments.items():
        src_vidname, dst_vidname = vidnames
        src_video = src.index.name_to_video[src_vidname]
        dst_video = dst.index.name_to_video[dst_vidname]

        if 0:
            # OVERSIMPLIFIED LOGIC
            # In most cases the source and destination videos will have the exact same
            # width / height, but in some cases they may differ if one is downsampled.
            # Just to be safe, assume the videos align and compute a scale factor
            # transform to account for this case (its just the identity if width/height
            # are the same).
            fx = dst_video['width'] / src_video['width']
            fy = dst_video['height'] / src_video['height']
            warp_dstvid_from_srcvid = kwimage.Affine.scale((fx, fy))
        else:
            # Better logic
            dst_crs = dst_video['wld_crs_info']
            src_crs = src_video['wld_crs_info']

            warp_src_from_wld = kwimage.Affine.coerce(src_video['warp_wld_to_vid'])
            warp_dst_from_wld = kwimage.Affine.coerce(dst_video['warp_wld_to_vid'])
            warp_wld_from_src = warp_src_from_wld.inv()

            warp_dstwld_from_srcwld = None

            if dst_crs != src_crs:
                msg = ub.codeblock(
                    f'''
                    Expected the same CRS but got:
                    dst_crs={dst_crs}
                    src_crs={src_crs}
                    ''')
                ALLOW_AFFINE_APPROXIMATE = config.allow_affine_approx
                if not ALLOW_AFFINE_APPROXIMATE:
                    raise AssertionError(msg)
                else:
                    rich.print('[yellow]WARNING: ' + msg)
                    rich.print('[yellow]WARNING: Estimating an approximate affine transform')
                    # We can construct a non-affine transform between the two CRS
                    # values, but because we are only writing metadata and not
                    # rewriting the pixel values, we cant use it. (kwcoco metadata
                    # only allows for affine transforms).
                    import pyproj
                    import numpy as np
                    # crs1 = pyproj.CRS.from_epsg(32639)
                    # crs2 = pyproj.CRS.from_epsg(32638)
                    crs1 = pyproj.CRS.from_authority(*src_crs['auth'])
                    crs2 = pyproj.CRS.from_authority(*dst_crs['auth'])
                    crs_tf = pyproj.Transformer.from_crs(crs_from=crs1, crs_to=crs2)

                    src_pxl_valid_region = kwimage.MultiPolygon.coerce(src_video['valid_region'])
                    src_wld_valid_region = src_pxl_valid_region.warp(warp_wld_from_src)
                    src_pts = np.concatenate([p.data['exterior'].data for p in src_wld_valid_region.data], axis=0)

                    # We now try be far too clever and estimate an affine
                    # approximation that gets does a good job in the region of
                    # the source image.
                    import numpy as np
                    xx1 = src_pts.T[0]
                    yy1 = src_pts.T[1]
                    xx2, yy2 = crs_tf.transform(xx1, yy1)
                    pts1 = np.stack([xx1, yy1], axis=1)
                    pts2 = np.stack([xx2, yy2], axis=1)
                    approx = kwimage.Affine.fit(pts1, pts2)
                    pts2_hat = kwimage.Points(xy=pts1).warp(approx)
                    error = np.abs(pts2_hat.xy - pts2)
                    max_error = error.max(axis=0)
                    ave_error = error.mean(axis=0)
                    print(f'ave_error={ave_error}')
                    print(f'max_error={max_error}')
                    warp_dstwld_from_srcwld = approx

            if warp_dstwld_from_srcwld is not None:
                # Using the hack
                warp_dstvid_from_srcvid = warp_dst_from_wld @ warp_dstwld_from_srcwld @ warp_wld_from_src
            else:
                warp_dstvid_from_srcvid = warp_dst_from_wld @ warp_wld_from_src

        for assignment in vid_assignments:
            src_image_id = assignment['src_image_id']
            dst_image_id = assignment['dst_image_id']
            src_coco_img = src.coco_image(src_image_id)
            dst_coco_img = dst.coco_image(dst_image_id)

            # Compute the alignment between assigned images.
            warp_srcvid_from_srcimg = src_coco_img.warp_vid_from_img
            warp_dstimg_from_dstvid = dst_coco_img.warp_img_from_vid
            warp_dstimg_from_srcimg = (
                warp_dstimg_from_dstvid @
                warp_dstvid_from_srcvid @
                warp_srcvid_from_srcimg
            )

            # For each asset in the src, check if we want to transfer it.
            img_assets_to_transfer = []
            for asset in src_coco_img.iter_asset_objs():
                asset_channels = kwcoco.FusedChannelSpec.coerce(asset['channels'])
                if asset_channels.intersection(channels_to_transfer):
                    img_assets_to_transfer.append(asset)

            for old_asset in img_assets_to_transfer:
                new_asset = old_asset.copy()
                # Ensure the filename points to the correct place (just use the
                # absolute path for simplicity for now)

                # Handle the case where the transform is not identity
                warp_srcimg_from_aux = kwimage.Affine.coerce(old_asset['warp_aux_to_img'])
                warp_dstimg_from_aux = warp_dstimg_from_srcimg @ warp_srcimg_from_aux
                new_asset['warp_aux_to_img'] = warp_dstimg_from_aux.concise()
                assets_to_transfer.append((dst_coco_img, old_asset, new_asset))

    print(f'Found {len(assets_to_transfer)} assets to tranfer')

    # Update the destination kwcoco file and prepare any copy jobs that need to
    # be done.
    src_bundle_dpath = src.bundle_dpath
    dst_bundle_dpath = dst.bundle_dpath
    copy_assets = config.copy_assets
    copy_tasks = []
    for dst_coco_img, src_asset, new_asset in assets_to_transfer:
        new_fname, copy_task = _make_new_asset_fname(
            src_asset, src_bundle_dpath, dst_bundle_dpath, copy_assets)
        if copy_task:
            copy_tasks.append(copy_task)

        new_asset['file_name'] = new_fname
        new_asset['image_id'] = dst_coco_img['id']
        dst_coco_img.add_asset(**new_asset)

    if not copy_assets:
        assert len(copy_tasks) == 0
        print('Copy assets is False. Only transfering reference to assets.')
    else:
        print(f'Found {len(copy_tasks)} assets to copy')

    rich.print(f'Dest Bundle: [link={dst.bundle_dpath}]{dst.bundle_dpath}[/link]')
    if copy_tasks:
        from kwutil import copy_manager
        copyman = copy_manager.CopyManager(workers=config.io_workers)
        for task in copy_tasks:
            copyman.submit(src=task['src'], dst=task['dst'],
                           overwrite=True)
        copyman.run(desc='Copy Assets')

    obj = proc_context.stop()
    dst.dataset['info'].append(obj)

    dst._ensure_json_serializable()
    if not do_return:
        print('Writing new coco file')
        dst.dump()
        print(f'Wrote transfered features to: {dst.fpath}')
    else:
        return dst


def _update_coco_fpath(self, new_fpath):
    # New method for more robustly updating the file path and bundle
    # directory, still a WIP
    if new_fpath is None:
        # Bundle directory is clobbered, so we should make everything
        # absolute
        self.reroot(absolute=True)
    else:
        old_fpath = self.fpath
        if old_fpath is not None:
            old_fpath_ = ub.Path(old_fpath)
            new_fpath_ = ub.Path(new_fpath)

            same_bundle = (
                (old_fpath_.parent == new_fpath_.parent) or
                (old_fpath_.resolve() == new_fpath_.resolve())
            )
            if not same_bundle:
                # The bundle directory has changed, so we need to reroot
                new_root = new_fpath_.parent
                self.reroot(new_root)

        self._fpath = new_fpath
        self._infer_dirs()


def _make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, copy_assets):
    """
    Find a new path for the transfered asset relative to the destination bundle.

    Args:
        src_asset (Dict): The asset dictionary we are transfering from src to dst
        src_bundle_dpath (str | PathLike): source bundle
        dst_bundle_dpath (str | PathLike): dest bundle
        copy_assets (bool): if True, build a new fname to copy to if possible

    Returns:
        Tuple[str, Dict | None]:
            the new fname and a dictionary containing the copy task
            if we need to perform one.

    TODO:
        port to kwcoco proper and move most of these tests to a unit test

    Example:
        >>> from geowatch.tasks.cold.transfer_features import _make_new_asset_fname

        >>> # Case: asset is absolute and inside src bundle
        >>> src_asset = {'file_name': '/my/src/rel/img.tif'}
        >>> src_bundle_dpath = '/my/src'
        >>> dst_bundle_dpath = '/my/dst'
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, False))
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, True))
        ('/my/src/rel/img.tif', None)
        ('rel/img.tif', {'src': '/my/src/rel/img.tif', 'dst': '/my/dst/rel/img.tif'})

        >>> # Case: asset is absolute and inside dst bundle
        >>> src_asset = {'file_name': '/my/dst/rel/img.tif'}
        >>> src_bundle_dpath = '/my/src'
        >>> dst_bundle_dpath = '/my/dst'
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, False))
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, True))
        ('rel/img.tif', None)
        ('rel/img.tif', None)

        >>> # Case: asset is absolute and inside neither bundle
        >>> src_asset = {'file_name': '/abs/img.tif'}
        >>> src_bundle_dpath = '/my/src'
        >>> dst_bundle_dpath = '/my/dst'
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, False))
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, True))
        ('/abs/img.tif', None)
        ('/abs/img.tif', None)

        >>> # Case: asset is relative and inside src bundle
        >>> src_asset = {'file_name': 'rel/img.tif'}
        >>> src_bundle_dpath = '/my/src'
        >>> dst_bundle_dpath = '/my/dst'
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, False))
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, True))
        ('/my/src/rel/img.tif', None)
        ('rel/img.tif', {'src': '/my/src/rel/img.tif', 'dst': '/my/dst/rel/img.tif'})

        >>> # Case: asset is relative and inside src bundle but with ..
        >>> src_asset = {'file_name': '../src/rel/img.tif'}
        >>> src_bundle_dpath = '/my/src'
        >>> dst_bundle_dpath = '/my/dst'
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, False))
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, True))
        ('/my/src/../src/rel/img.tif', None)
        ('rel/img.tif', {'src': '/my/src/../src/rel/img.tif', 'dst': '/my/dst/rel/img.tif'})

        >>> # Case: asset is relative and inside dst bundle
        >>> src_asset = {'file_name': '../dst/img.tif'}
        >>> src_bundle_dpath = '/my/src'
        >>> dst_bundle_dpath = '/my/dst'
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, False))
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, True))
        ('img.tif', None)
        ('img.tif', None)

        >>> # Case: asset is relative and inside neither bundle
        >>> src_asset = {'file_name': '../neither/img.tif'}
        >>> src_bundle_dpath = '/my/src'
        >>> dst_bundle_dpath = '/my/dst'
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, False))
        >>> print(_make_new_asset_fname(src_asset, src_bundle_dpath, dst_bundle_dpath, True))
        ('/my/src/../neither/img.tif', None)
        ('/my/neither/img.tif', None)
    """
    import warnings
    old_asset_fpath = join(src_bundle_dpath, src_asset['file_name'])
    dst_rel_old_asset_fpath = os.path.relpath(old_asset_fpath, dst_bundle_dpath)
    old_points_outside_dst_bundle = ('..' in ub.Path(dst_rel_old_asset_fpath).parts)

    copy_task = None

    if copy_assets:
        src_rel_old_asset_fpath = os.path.relpath(old_asset_fpath, src_bundle_dpath)
        old_points_outside_src_bundle = ('..' in ub.Path(src_rel_old_asset_fpath).parts)
        if not old_points_outside_dst_bundle:
            warnings.warn('An asset to copy already points inside of the dst bundle, not copying')
            new_fname = dst_rel_old_asset_fpath
        elif old_points_outside_src_bundle:
            # Dont want to deal with the corner case of figuring out a
            # good internal-to-dst-bundle location for an asset that
            # doesnt have a defined internal-to-src-bundle location.
            warnings.warn('An asset to copy points outside of the src bundle, not copying')
            new_fname = os.path.normpath(old_asset_fpath)
        else:
            # This case is safer to copy
            new_fname = src_rel_old_asset_fpath
            copy_task = {
                'src': old_asset_fpath,
                'dst': join(dst_bundle_dpath, new_fname),
            }
    else:
        if old_points_outside_dst_bundle:
            # The asset is outside the dst bundle, so we need to point at
            # the absolute path to it.
            new_fname = old_asset_fpath
        else:
            # Not copying, but safe to use relative paths because the asset
            # is already inside the destination bundle.
            new_fname = dst_rel_old_asset_fpath

    return new_fname, copy_task


def _test_cases():
    """
    Embedding test cases as doctests for now

    Example:
        >>> # Test case: transfer from temporal-lores to temporal-hires
        >>> from geowatch.tasks.cold.transfer_features import transfer_features_main
        >>> from geowatch.tasks.cold.transfer_features import *
        >>> import geowatch
        >>> dset1 = geowatch.coerce_kwcoco('geowatch-msi', geodata=True, heatmap=True, dates=True)
        >>> dset2 = dset1.copy()
        >>> # Remove half of the images from dset1
        >>> dset1.remove_images(list(dset1.images())[::2])
        >>> # Remove saliency assets from dset2
        >>> for img in dset2.images().coco_images:
        >>>     asset = img.find_asset_obj('salient')
        >>>     img['auxiliary'].remove(asset)
        >>> dset2_orig = dset2.copy()
        >>> # Transfer the saliency from dset1 onto dset2
        >>> kwargs = TransferCocoConfig(**{
        >>>     'src_kwcoco': dset1,
        >>>     'dst_kwcoco': dset2,
        >>>     'new_coco_fpath': 'return',
        >>>     'channels_to_transfer': 'salient',
        >>>     'copy_assets': False,
        >>> })
        >>> cmdline = False
        >>> new_dset = transfer_features_main(cmdline, **kwargs)
        >>> assert new_dset is dset2, 'modifies combine_fpath inplace'
        >>> from geowatch.utils import kwcoco_extensions
        >>> stats1 = kwcoco_extensions.coco_channel_stats(dset1)
        >>> stats2 = kwcoco_extensions.coco_channel_stats(dset2)
        >>> stats2_orig = kwcoco_extensions.coco_channel_stats(dset2_orig)
        >>> assert stats2['single_chan_hist']['salient'] == stats1['single_chan_hist']['salient'], 'all assets should transfer'
        >>> assert stats2['single_chan_hist']['salient'] == dset2.n_images // 2, 'only some dst images get saliency'
        >>> assert stats2_orig['single_chan_hist']['salient'] == 0

    Example:
        >>> # Test case: transfer from temporal-lores to temporal-hires multiple transfer
        >>> from geowatch.tasks.cold.transfer_features import transfer_features_main
        >>> from geowatch.tasks.cold.transfer_features import *
        >>> import geowatch
        >>> dset1 = geowatch.coerce_kwcoco('geowatch-msi', geodata=True, heatmap=True, dates=True)
        >>> dset2 = dset1.copy()
        >>> # Remove half of the images from dset1
        >>> dset1.remove_images(list(dset1.images())[::2])
        >>> # Remove saliency assets from dset2
        >>> for img in dset2.images().coco_images:
        >>>     asset = img.find_asset_obj('salient')
        >>>     img['auxiliary'].remove(asset)
        >>> dset2_orig = dset2.copy()
        >>> # Transfer the saliency from dset1 onto dset2
        >>> kwargs = TransferCocoConfig(**{
        >>>     'src_kwcoco': dset1,
        >>>     'dst_kwcoco': dset2,
        >>>     'new_coco_fpath': 'return',
        >>>     'respect_sensors': False,
        >>>     'max_propogate': None,
        >>>     'channels_to_transfer': 'salient',
        >>>     'copy_assets': False,
        >>> })
        >>> cmdline = False
        >>> new_dset = transfer_features_main(cmdline, **kwargs)
        >>> assert new_dset is dset2, 'modifies combine_fpath inplace'
        >>> from geowatch.utils import kwcoco_extensions
        >>> stats1 = kwcoco_extensions.coco_channel_stats(dset1)
        >>> stats2 = kwcoco_extensions.coco_channel_stats(dset2)
        >>> stats2_orig = kwcoco_extensions.coco_channel_stats(dset2_orig)
        >>> assert stats2['single_chan_hist']['salient'] > stats1['single_chan_hist']['salient'], 'some assets should transfer multiple times'
        >>> assert stats2_orig['single_chan_hist']['salient'] == 0

    Example:
        >>> # Test case: transfer from temporal-hires to temporal-lores
        >>> from geowatch.tasks.cold.transfer_features import transfer_features_main
        >>> from geowatch.tasks.cold.transfer_features import *
        >>> import geowatch
        >>> dset1 = geowatch.coerce_kwcoco('geowatch-msi', geodata=True, heatmap=True, dates=True)
        >>> dset2 = dset1.copy()
        >>> # Remove half of the images from dset2
        >>> dset2.remove_images(list(dset1.images())[::2])
        >>> # Remove saliency assets from dset2
        >>> for img in dset2.images().coco_images:
        >>>     asset = img.find_asset_obj('salient')
        >>>     img['auxiliary'].remove(asset)
        >>> dset2_orig = dset2.copy()
        >>> # Transfer the saliency from dset1 onto dset2
        >>> kwargs = TransferCocoConfig(**{
        >>>     'src_kwcoco': dset1,
        >>>     'dst_kwcoco': dset2,
        >>>     'new_coco_fpath': 'return',
        >>>     'channels_to_transfer': 'salient',
        >>>     'copy_assets': False,
        >>> })
        >>> cmdline = False
        >>> new_dset = transfer_features_main(cmdline, **kwargs)
        >>> assert new_dset is dset2, 'modifies combine_fpath inplace'
        >>> from geowatch.utils import kwcoco_extensions
        >>> stats1 = kwcoco_extensions.coco_channel_stats(dset1)
        >>> stats2 = kwcoco_extensions.coco_channel_stats(dset2)
        >>> stats2_orig = kwcoco_extensions.coco_channel_stats(dset2_orig)
        >>> assert stats2['single_chan_hist']['salient'] < stats1['single_chan_hist']['salient'], 'more saliency in dset1 than dset2'
        >>> assert stats2['single_chan_hist']['salient'] == dset2.n_images, 'all dst images get saliency'
        >>> assert stats2_orig['single_chan_hist']['salient'] == 0
    """


if __name__ == '__main__':
    transfer_features_main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains blocks used in MOAT [1].

This file includes MOAT blocks [1] and MBConv [2] blocks.
The MBConv implementation is based on CoAtNet [3].

[1] MOAT: Alternating Mobile Convolution and Attention
    Brings Strong Vision Models,
    arXiv: 2210.01820.
      Chenglin Yang, Siyuan Qiao, Qihang Yu, Xiaoding Yuan,
      Yukun Zhu, Alan Yuille, Hartwig Adam, Liang-Chieh Chen.

[2] MobileNetV2: Inverted Residuals and Linear Bottlenecks
    CVPR 2018.
      Mark Sandler, Andrew Howard, Menglong Zhu,
      Andrey Zhmoginov, Liang-Chieh Chen.

[3] CoAtNet: Marrying Convolution and Attention for All Data Sizes
    NeurIPS 2021.
      Zihang Dai, Hanxiao Liu, Quoc V. Le, Mingxing Tan.
"""

import math

import tensorflow as tf
from deeplab2.model.layers.moat_attention import Attention
from deeplab2.utils.hparam_configs import create_config_from_dict


def drop_connect(inputs: tf.Tensor, training: bool,
                 survival_prob: float) -> tf.Tensor:
    """Drops the entire conv with given survival probability [1].

    [1] Deep Networks with Stochastic Depth,
        ECCV 2016.
          Gao Huang, Yu Sun, Zhuang Liu, Daniel Sedra, Kilian Q. Weinberger.

    Args:
      inputs: A tensor with shape [batch_size, height, width, channels].
      training: A boolen, whether in training mode or not.
      survival_prob: A float, 1 - drop_path_rate [1].

    Returns:
      output: A tensor with shape [batch_size, height, width, channels]
    """

    if not training:
        return inputs
    batch_size = tf.shape(inputs)[0]
    random_tensor = survival_prob
    random_tensor += tf.random.uniform([batch_size], dtype=inputs.dtype)
    for _ in range(inputs.shape.rank - 1):
        random_tensor = tf.expand_dims(random_tensor, axis=-1)
    binary_tensor = tf.floor(random_tensor)
    # Unlike the conventional way that we multiply survival_prob at test time, we
    # divide survival_prob at training time, so no additional compute is needed at
    # test time.
    output = inputs / survival_prob * binary_tensor
    return output


def residual_add_with_drop_path(
        residual: tf.Tensor, shortcut: tf.Tensor,
        survival_prob: float, training: bool) -> tf.Tensor:
    """Combines residual and shortcut."""
    if survival_prob is not None and 0 < survival_prob < 1:
        residual = drop_connect(residual, training, survival_prob)
    return shortcut + residual


class SqueezeAndExcitation(tf.keras.layers.Layer):
    """Implementation of Squeeze-and-excitation layer."""

    def _retrieve_config(self, config):
        """Retrieves the config of SqueezeAndExcitation.

        Args:
          config: A dictionary containing the following keys.
            -se_filters: An integer, feature channels of bottlneck
              in SqueezeAndExcitation.
            -output_filters: An integer, output channels.
            -activation: Activation layer class.
            -survival_prob: A float, 1 - drop_path_rate.
            -kernel_initializer: Initializer for the kernel weights matrix.
            -bias_initializer: Initializer for the bias vector.
            -name: A string, layer name.

        Returns:
          A hparam_configs.
        """

        required_keys = ['se_filters', 'output_filters']
        optional_keys = {
            'activation': tf.keras.activations.swish,
            'kernel_initializer': tf.random_normal_initializer(stddev=0.02),
            'bias_initializer': tf.zeros_initializer,
            'name': 'se',
        }
        config = create_config_from_dict(config, required_keys, optional_keys)
        return config

    def __init__(self, **config):
        self._config = self._retrieve_config(config)
        super().__init__(name=self._config.name)

        self._se_reduce = tf.keras.layers.Conv2D(
            self._config.se_filters,
            kernel_size=1,
            strides=1,
            padding='same',
            use_bias=True,
            kernel_initializer=self._config.kernel_initializer,
            bias_initializer=self._config.bias_initializer,
            name='reduce_conv2d')
        self._se_expand = tf.keras.layers.Conv2D(
            self._config.output_filters,
            kernel_size=1,
            strides=1,
            padding='same',
            use_bias=True,
            kernel_initializer=self._config.kernel_initializer,
            bias_initializer=self._config.bias_initializer,
            name='expand_conv2d')
        self.activation_fn = self._config.activation

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        _ = inputs.get_shape().with_rank(4)
        se_tensor = tf.reduce_mean(inputs, [1, 2], keepdims=True)
        se_tensor = self._se_expand(self.activation_fn(self._se_reduce(se_tensor)))
        return tf.sigmoid(se_tensor) * inputs


class MBConvBlock(tf.keras.layers.Layer):
    """Implementation of inverted bottleneck block [1].

    [1] MobileNetV2: Inverted Residuals and Linear Bottlenecks
      CVPR 2018.
        Mark Sandler, Andrew Howard, Menglong Zhu,
        Andrey Zhmoginov, Liang-Chieh Chen.
    """

    def _retrieve_config(self, config):
        """Retrieves the config of MBConvBlock.

        Args:
          config: A dictionary containing the following keys.
            -hidden_size: An integer, output channel.
            -kernel_size: An integer, kernel size of MBConv
            -expansion_rate: An integer, expansion rate of MBConv.
            -se_ratio: An integer, expansion ratio of SqueezeAndExcitation
              in MBConv.
            -block_stride: An integer, stride of MBConv.
            -pool_size: An integer, kernel size for pooling in shortcut branch.
              For classification, pool size 2x2 saves model flops.
              For downstream tasks, pool size 3x3 is preferred for better feature
              alignments.
            -norm_class: Normalization layer class.
            -activation: Activation layer class.
            -survival_prob: A float, 1 - drop_path_rate.
            -kernel_initializer: Initializer for the kernel weights matrix.
            -bias_initializer: Initializer for the bias vector.
            -name: A string, layer name.

        Returns:
          A Config class: hparams_config.Config.
        """

        required_keys = ['hidden_size']
        optional_keys = {
            'kernel_size': 3,
            'expansion_rate': 4,
            'se_ratio': 0.25,
            'block_stride': 1,
            'pool_size': 2,
            'norm_class': tf.keras.layers.experimental.SyncBatchNormalization,
            'activation': tf.keras.activations.gelu,
            'survival_prob': None,
            'kernel_initializer': tf.random_normal_initializer(stddev=0.02),
            'bias_initializer': tf.zeros_initializer,
            'name': 'mbconv',
        }
        config = create_config_from_dict(config, required_keys, optional_keys)
        return config

    def __init__(self, **config):
        self._config = self._retrieve_config(config)
        super().__init__(name=self._config.name)
        self._activation_fn = self._config.activation
        self._norm_class = self._config.norm_class

    def build(self, input_shape) -> None:
        input_size = input_shape[-1]
        inner_size = self._config.hidden_size * self._config.expansion_rate

        self._shortcut_conv = None
        if input_size != self._config.hidden_size:
            self._shortcut_conv = tf.keras.layers.Conv2D(
                filters=self._config.hidden_size,
                kernel_size=1,
                strides=1,
                padding='same',
                kernel_initializer=self._config.kernel_initializer,
                bias_initializer=self._config.bias_initializer,
                use_bias=True,
                name='shortcut_conv')

        self._pre_norm = self._norm_class(name='pre_norm')
        self._expand_conv = tf.keras.layers.Conv2D(
            filters=inner_size,
            kernel_size=1,
            strides=1,
            kernel_initializer=self._config.kernel_initializer,
            padding='same',
            use_bias=False,
            name='expand_conv')
        self._expand_norm = self._norm_class(name='expand_norm')
        self._depthwise_conv = tf.keras.layers.DepthwiseConv2D(
            kernel_size=self._config.kernel_size,
            strides=self._config.block_stride,
            depthwise_initializer=self._config.kernel_initializer,
            padding='same',
            use_bias=False,
            name='depthwise_conv')
        self._depthwise_norm = self._norm_class(name='depthwise_norm')

        self._se = None
        if self._config.se_ratio is not None:
            se_filters = max(1, int(self._config.hidden_size * self._config.se_ratio))
            self._se = SqueezeAndExcitation(
                se_filters=se_filters,
                output_filters=inner_size,
                kernel_initializer=self._config.kernel_initializer,
                bias_initializer=self._config.bias_initializer,
                name='se')

        self._shrink_conv = tf.keras.layers.Conv2D(
            filters=self._config.hidden_size,
            kernel_size=1,
            strides=1,
            padding='same',
            kernel_initializer=self._config.kernel_initializer,
            bias_initializer=self._config.bias_initializer,
            use_bias=True,
            name='shrink_conv')

    def _shortcut_downsample(self, inputs, name):
        output = inputs
        if self._config.block_stride > 1:
            pooling_layer = tf.keras.layers.AveragePooling2D(
                pool_size=self._config.pool_size,
                strides=self._config.block_stride,
                padding='same',
                name=name,
            )
            if output.dtype == tf.float32:
                output = pooling_layer(output)
            else:
                # We find that in our code base, the output dtype of pooling is float32
                # no matter whether its input and compute dtype is bfloat16 or
                # float32. So we explicitly cast the output dtype of pooling to be the
                # model compute dtype.
                output = tf.cast(pooling_layer(
                    tf.cast(output, tf.float32)), output.dtype)
        return output

    def _shortcut_branch(self, inputs):
        shortcut = self._shortcut_downsample(inputs, name='shortcut_pool')
        if self._shortcut_conv:
            shortcut = self._shortcut_conv(shortcut)
        return shortcut

    def call(self, inputs: tf.Tensor, training: bool) -> tf.Tensor:
        shortcut = self._shortcut_branch(inputs)
        output = self._pre_norm(inputs, training=training)
        output = self._expand_conv(output)
        output = self._expand_norm(output, training=training)
        output = self._activation_fn(output)
        output = self._depthwise_conv(output)
        output = self._depthwise_norm(output, training=training)
        output = self._activation_fn(output)
        if self._se:
            output = self._se(output)
        output = self._shrink_conv(output)
        output = residual_add_with_drop_path(
            output, shortcut,
            self._config.survival_prob, training)
        return output


class MOATBlock(tf.keras.layers.Layer):
    """Implementation of MOAT block [1].

    [1] MOAT: Alternating Mobile Convolution and Attention
      Brings Strong Vision Models,
      arXiv: ----.----.
        Chenglin Yang, Siyuan Qiao, Qihang Yu, Xiaoding Yuan,
        Yukun Zhu, Alan Yuille, Hartwig Adam, Liang-Chieh Chen.
    """

    def _retrieve_config(self, config):
        """Retrieves the config of MOATBlock.

        Args:
          config: A dictionary containing the following keys.
            -hidden_size: An integer, output channels.
            -kernel_size: An integer, kernel size of MBConv
            -expansion_rate: An integer, expansion rate of MBConv.
            -block_stride: An integer, stride of MBConv.
            -pool_size: An integer, kernel size for pooling in shortcut branch.
              For classification, pool size 2x2 saves model flops.
              For downstream tasks, pool size 3x3 is preferred for better feature
              alignments.
            -norm_class: Normalization layer class in MBConv.
            -activation: Activation layer class in MBConv.
            -head_size: An integer, feature channels per head in Attention.
            -window_size: A list of two integers, spatial size of input for window
              attention, [height, width]. If None, global attention will be enabled.
            -relative_position_embedding_type: A string, type of relative position
              embedding in Attention. If None, no relative position embedding will
              be used.
            -position_embedding_size: An integer, specifying the position embedding
             size. If the feature map have larger size, the position embedding will
             be interpolated to match it.
            -ln_epsilon: A float, epsilon for layer normalization in Attention.
            -survival_prob: A float, 1 - drop_path_rate.
            -kernel_initializer: Initializer for the kernel weights matrix.
            -bias_initializer: Initializer for the bias vector.
            -use_checkpointing_for_attention: A boolean, specifying whether to use
              checkpointing for attention.
            -name: A string, layer name.

        Returns:
          A Config class: hparams_config.Config.

        Raises:
          ValueError: If the window size is not None and the length of window_size
            is not two.
        """

        required_keys = ['hidden_size']
        optional_keys = {
            'kernel_size': 3,
            'expansion_rate': 4,
            'block_stride': 2,
            'pool_size': 2,
            'norm_class': tf.keras.layers.experimental.SyncBatchNormalization,
            'activation': tf.keras.activations.gelu,
            'head_size': 32,
            'window_size': None,
            'relative_position_embedding_type': '2d_multi_head',
            'position_embedding_size': 7,
            'ln_epsilon': 1e-5,
            'survival_prob': None,
            'kernel_initializer': tf.random_normal_initializer(stddev=0.02),
            'bias_initializer': tf.zeros_initializer,
            'use_checkpointing_for_attention': False,
            'name': 'moat',
        }
        config = create_config_from_dict(config, required_keys, optional_keys)
        return config

    def __init__(self, **config):
        self._config = self._retrieve_config(config)
        super().__init__(name=self._config.name)
        self._activation_fn = self._config.activation
        self._norm_class = self._config.norm_class

    def build(self, input_shape) -> None:
        height, width, input_size = input_shape[-3:]
        inner_size = self._config.hidden_size * self._config.expansion_rate

        if self._config.window_size:
            if (isinstance(self._config.window_size, list) and len(self._config.window_size)) == 2:
                self._window_height = self._config.window_size[0]
                self._window_width = self._config.window_size[1]
            else:
                raise ValueError((
                    'The window size should be a list of two ints',
                    '[height, width], if specified.'))
        else:
            self._window_height = math.ceil(float(height) / self._config.block_stride)
            self._window_width = math.ceil(float(width) / self._config.block_stride)

        self._shortcut_conv = None
        if input_size != self._config.hidden_size:
            self._shortcut_conv = tf.keras.layers.Conv2D(
                filters=self._config.hidden_size,
                kernel_size=1,
                strides=1,
                padding='same',
                kernel_initializer=self._config.kernel_initializer,
                bias_initializer=self._config.bias_initializer,
                use_bias=True,
                name='shortcut_conv')

        self._pre_norm = self._norm_class(name='pre_norm')
        self._expand_conv = tf.keras.layers.Conv2D(
            filters=inner_size,
            kernel_size=1,
            strides=1,
            kernel_initializer=self._config.kernel_initializer,
            padding='same',
            use_bias=False,
            name='expand_conv')
        self._expand_norm = self._norm_class(name='expand_norm')
        self._depthwise_conv = tf.keras.layers.DepthwiseConv2D(
            kernel_size=self._config.kernel_size,
            strides=self._config.block_stride,
            depthwise_initializer=self._config.kernel_initializer,
            padding='same',
            use_bias=False,
            name='depthwise_conv')
        self._depthwise_norm = self._norm_class(name='depthwise_norm')
        self._shrink_conv = tf.keras.layers.Conv2D(
            filters=self._config.hidden_size,
            kernel_size=1,
            strides=1,
            padding='same',
            kernel_initializer=self._config.kernel_initializer,
            bias_initializer=self._config.bias_initializer,
            use_bias=True,
            name='shrink_conv')

        self._attention_norm = tf.keras.layers.LayerNormalization(
            axis=-1,
            epsilon=self._config.ln_epsilon,
            name='attention_norm')

        scale_ratio = None
        if self._config.relative_position_embedding_type:
            if self._config.position_embedding_size is None:
                raise ValueError(
                    'The position embedding size need to be specified ' +
                    'if relative position embedding is used.')
            scale_ratio = [
                self._window_height / self._config.position_embedding_size,
                self._window_width / self._config.position_embedding_size,
            ]

        self._attention = Attention(
            hidden_size=self._config.hidden_size,
            head_size=self._config.head_size,
            relative_position_embedding_type=(
                self._config.relative_position_embedding_type),
            scale_ratio=scale_ratio,
            kernel_initializer=self._config.kernel_initializer,
            bias_initializer=self._config.bias_initializer)

    def _make_wondows(self, inputs):
        _, height, width, channels = inputs.get_shape().with_rank(4).as_list()
        inputs = tf.reshape(
            inputs,
            (-1,
             height // self._window_height, self._window_height,
             width // self._window_width, self._window_width,
             channels))
        inputs = tf.transpose(inputs, (0, 1, 3, 2, 4, 5))
        inputs = tf.reshape(
            inputs,
            (-1, self._window_height, self._window_width, channels))
        return inputs

    def _remove_windows(self, inputs, height, width):
        _, _, channels = inputs.get_shape().with_rank(3).as_list()
        inputs = tf.reshape(inputs, [
            -1, height // self._window_height, width // self._window_width,
            self._window_height, self._window_width, channels
        ])
        inputs = tf.transpose(inputs, (0, 1, 3, 2, 4, 5))
        inputs = tf.reshape(inputs, (-1, height, width, channels))
        return inputs

    def _shortcut_downsample(self, inputs, name):
        output = inputs
        if self._config.block_stride > 1:
            pooling_layer = tf.keras.layers.AveragePooling2D(
                pool_size=self._config.pool_size,
                strides=self._config.block_stride,
                padding='same',
                name=name,
            )
            if output.dtype == tf.float32:
                output = pooling_layer(output)
            else:
                # We find that in our code base, the output dtype of pooling is float32
                # no matter whether its input and compute dtype is bfloat16 or
                # float32. So we explicitly cast the output dtype of pooling to be the
                # model compute dtype.
                output = tf.cast(pooling_layer(
                    tf.cast(output, tf.float32)), output.dtype)
        return output

    def _shortcut_branch(self, inputs):
        shortcut = self._shortcut_downsample(inputs, name='shortcut_pool')
        if self._shortcut_conv:
            shortcut = self._shortcut_conv(shortcut)
        return shortcut

    def call(self, inputs: tf.Tensor, training: bool) -> tf.Tensor:
        mbconv_shortcut = self._shortcut_branch(inputs)
        output = self._pre_norm(inputs, training=training)
        output = self._expand_conv(output)
        output = self._expand_norm(output, training=training)
        output = self._activation_fn(output)
        output = self._depthwise_conv(output)
        output = self._depthwise_norm(output, training=training)
        output = self._activation_fn(output)
        output = self._shrink_conv(output)
        output = residual_add_with_drop_path(
            output, mbconv_shortcut,
            self._config.survival_prob, training)

        # For classification, the window size is the same as feature map size.
        # For downstream tasks, the window size can be set the same as
        # classification's.
        attention_shortcut = output

        def _func(output):
            output = self._attention_norm(output)
            _, height, width, _ = output.get_shape().with_rank(4).as_list()
            output = self._make_wondows(output)
            output = self._attention(output)
            output = self._remove_windows(output, height, width)
            return output

        func = _func
        if self._config.use_checkpointing_for_attention:
            func = tf.recompute_grad(_func)

        output = func(output)
        output = residual_add_with_drop_path(
            output, attention_shortcut,
            self._config.survival_prob, training)
        return output



"""Simple creation of command line interfaces."""

import inspect
from typing import Any, Callable, Dict, List, Optional, Type, Union

from ._actions import ActionConfigFile, _ActionPrintConfig, remove_actions
from ._core import ArgumentParser
from ._deprecated import deprecation_warning_cli_return_parser
from ._namespace import Namespace, dict_to_namespace
from ._optionals import get_doc_short_description
from ._util import default_config_option_help

__all__ = ["CLI"]


ComponentType = Union[Callable, Type]
DictComponentsType = Dict[str, Union[ComponentType, "DictComponentsType"]]
ComponentsType = Optional[Union[ComponentType, List[ComponentType], DictComponentsType]]


def CLI(
    components: ComponentsType = None,
    args: Optional[List[str]] = None,
    config_help: str = default_config_option_help,
    set_defaults: Optional[Dict[str, Any]] = None,
    as_positional: bool = True,
    fail_untyped: bool = True,
    parser_class: Type[ArgumentParser] = ArgumentParser,
    **kwargs,
):
    """Simple creation of command line interfaces.

    Creates an argument parser from one or more functions/classes, parses
    arguments and runs one of the functions or class methods depending on what
    was parsed. If the 'components' parameter is not given, then the components
    will be all the locals in the context and defined in the same module as from
    where CLI is called.

    Args:
        components: One or more functions/classes to include in the command line interface.
        args: List of arguments to parse or None to use sys.argv.
        config_help: Help string for config file option in help.
        set_defaults: Dictionary of values to override components defaults.
        as_positional: Whether to add required parameters as positional arguments.
        fail_untyped: Whether to raise exception if a required parameter does not have a type.
        parser_class: The ArgumentParser class to use.
        **kwargs: Used to instantiate :class:`.ArgumentParser`.

    Returns:
        The value returned by the executed function or class method.
    """
    return_parser = kwargs.pop("return_parser", False)
    caller = inspect.stack()[1][0]

    if components is None:
        module = inspect.getmodule(caller).__name__  # type: ignore[union-attr]
        components = [
            v
            for v in caller.f_locals.values()
            if ((inspect.isclass(v) or callable(v)) and getattr(inspect.getmodule(v), "__name__", None) == module)
        ]
        if len(components) == 0:
            raise ValueError(
                "Either components parameter must be given or there must be at least one "
                "function or class among the locals in the context where CLI is called."
            )

    if isinstance(components, list) and len(components) == 1:
        components = components[0]

    elif not components:
        raise ValueError("components parameter expected to be non-empty")

    if isinstance(components, list):
        unexpected = [c for c in components if not (inspect.isclass(c) or callable(c))]
    elif isinstance(components, dict):
        ns = dict_to_namespace(components)
        unexpected = [c for k, c in ns.items() if not k.endswith("._help") and not (inspect.isclass(c) or callable(c))]
    else:
        unexpected = [c for c in [components] if not (inspect.isclass(c) or callable(c))]
    if unexpected:
        raise ValueError(f"Unexpected components, not class or function: {unexpected}")

    parser = parser_class(default_meta=False, **kwargs)
    parser.add_argument("--config", action=ActionConfigFile, help=config_help)

    if not isinstance(components, (list, dict)):
        _add_component_to_parser(components, parser, as_positional, fail_untyped, config_help)
        if set_defaults is not None:
            parser.set_defaults(set_defaults)
        if return_parser:
            deprecation_warning_cli_return_parser()
            return parser
        cfg = parser.parse_args(args)
        cfg_init = parser.instantiate_classes(cfg)
        return _run_component(components, cfg_init)

    elif isinstance(components, list):
        components = {c.__name__: c for c in components}

    _add_subcommands(components, parser, config_help, as_positional, fail_untyped)

    if set_defaults is not None:
        parser.set_defaults(set_defaults)
    if return_parser:
        deprecation_warning_cli_return_parser()
        return parser
    cfg = parser.parse_args(args)
    init = parser.instantiate_classes(cfg)
    components_ns = dict_to_namespace(components)
    subcommand = init.get("subcommand")
    while isinstance(init.get(subcommand), Namespace) and isinstance(init[subcommand].get("subcommand"), str):
        subsubcommand = subcommand + "." + init[subcommand].get("subcommand")
        if subsubcommand in components_ns:
            subcommand = subsubcommand
        else:
            break
    component = components_ns[subcommand]
    return _run_component(component, init.get(subcommand))


def get_help_str(component, logger):
    if isinstance(component, dict):
        return component.get("_help")
    help_str = get_doc_short_description(component, logger=logger)
    if not help_str:
        help_str = str(component)
    return help_str


def _add_subcommands(
    components,
    parser: ArgumentParser,
    config_help: str,
    as_positional: bool,
    fail_untyped: bool,
) -> None:
    subcommands = parser.add_subcommands(required=True)
    for name, component in components.items():
        if name == "_help":
            continue
        description = get_help_str(component, parser.logger)
        subparser = type(parser)(description=description)
        subparser.add_argument("--config", action=ActionConfigFile, help=config_help)
        subcommands.add_subcommand(name, subparser, help=description)
        if isinstance(component, dict):
            _add_subcommands(component, subparser, config_help, as_positional, fail_untyped)
        else:
            added_args = _add_component_to_parser(component, subparser, as_positional, fail_untyped, config_help)
            if not added_args:
                remove_actions(subparser, (ActionConfigFile, _ActionPrintConfig))


def has_parameter(component, name):
    return name in inspect.signature(component).parameters.keys()


def _add_component_to_parser(
    component,
    parser: ArgumentParser,
    as_positional: bool,
    fail_untyped: bool,
    config_help: str,
):
    kwargs: dict = dict(as_positional=as_positional, fail_untyped=fail_untyped, sub_configs=True)
    if inspect.isclass(component):
        class_methods = [k for k, v in inspect.getmembers(component) if callable(v) and k[0] != "_"]
        if not class_methods:
            added_args = parser.add_class_arguments(component, as_group=False, **kwargs)
            if not parser.description:
                parser.description = get_help_str(component, parser.logger)
            return added_args
        added_args = parser.add_class_arguments(component, **kwargs)
        subcommands = parser.add_subcommands(required=True)
        for method in class_methods:
            method_object = getattr(component, method)
            description = get_help_str(method_object, parser.logger)
            subparser = type(parser)(description=description)
            if not has_parameter(getattr(component, method), "config"):
                subparser.add_argument("--config", action=ActionConfigFile, help=config_help)
            added_subargs = subparser.add_method_arguments(component, method, as_group=False, **kwargs)
            added_args += [f"{method}.{a}" for a in added_subargs]
            if not added_subargs:
                remove_actions(subparser, (ActionConfigFile, _ActionPrintConfig))
            subcommands.add_subcommand(method, subparser, help=get_help_str(method_object, parser.logger))
    else:
        added_args = parser.add_function_arguments(component, as_group=False, **kwargs)
        if not parser.description:
            parser.description = get_help_str(component, parser.logger)
    return added_args


def _run_component(component, cfg):
    cfg.pop("config", None)
    if not inspect.isclass(component):
        return component(**cfg)
    subcommand = cfg.pop("subcommand")
    if not subcommand:
        return component(**cfg)
    subcommand_cfg = cfg.pop(subcommand, {})
    subcommand_cfg.pop("config", None)
    component_obj = component(**cfg)
    return getattr(component_obj, subcommand)(**subcommand_cfg)



"""
Functions that may be moved to kwutil
"""
import itertools as it


# Backwards compatible variant used in geowatch util
def distributed_subitems(items, max_num=None):
    """
    Return a subset of items maximally distributed over the input index space.
    I.e. the chosen indexes maximize the space between them.

    Args:
        items (List | Dict): an ordered indexable object

    Returns:
        List | Dict: a subset of the input with a length at most ``max_num``.

    Note:
        Prefer using the generator variant :func:`farthest_first` instead.

    TODO:
        - [X] Find a better name. ChatGPT suggests using "spread", which I
              like. Maybe spreadsort, spreadshuffle, spredtraverse?
              spreadtake, takespread?
        - [ ] Figure out where this lives, probably kwutil.
        - [X] maybe we should force this to be a generator? Or make a generator variant?


    Example:
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> items = list(range(100))
        >>> max_num = 5
        >>> sub_items = distributed_subitems(items, max_num)
        >>> print(sub_items)
        [0, 25, 50, 75, 99]

    Example:
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> items = {chr(i): i for i in range(ord('a'), ord('a') + 26)}
        >>> max_num = 5
        >>> sub_items = distributed_subitems(items, max_num)
        >>> print(sub_items)
        {'a': 97, 'g': 103, 'n': 110, 't': 116, 'z': 122}
    """
    sub_index_gen = _farthest_first_indices(0, len(items))
    sub_indices = sorted(it.islice(sub_index_gen, max_num))
    if isinstance(items, dict):
        item_keys = list(items.keys())
        sub_keys = [item_keys[idx] for idx in sub_indices]
        sub_items = {key: items[key] for key in sub_keys}
    else:
        sub_items = [items[idx] for idx in sub_indices]
    return sub_items


def farthest_first(items, max_num=None, first=None):
    """
    Return a subset of items maximally distributed over the input index space.
    I.e. the chosen indexes maximize the space between them.

    Args:
        items (List | Dict): an ordered indexable object
        first (int | None): if specified, start with this index.

    Returns:
        List | Dict: a subset of the input with a length at most ``max_num``.

    Note:
        A farthest first sequence is not unique in general. The particular
        order this function returns may change in the future.

    Example:
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> items = list(range(100))
        >>> max_num = 5
        >>> sub_items = list(farthest_first(items, max_num))
        >>> print(list(sub_items))
        [99, 0, 50, 25, 75]

    Example:
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> items = {chr(i): i for i in range(ord('a'), ord('a') + 26)}
        >>> max_num = 5
        >>> sub_items = list(farthest_first(items, max_num))
        >>> print(dict(sub_items))
        {'z': 122, 'a': 97, 'n': 110, 'g': 103, 't': 116}

    Example:
        >>> # Choose a custom starting location
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> import ubelt as ub
        >>> items = list(range(10))
        >>> results = []
        >>> start, stop = 0, 10
        >>> for first in range(start, stop):
        >>>     results += [list(farthest_first(items, first=first))]
        >>> print(f'results = {ub.urepr(results, nl=1)}')
        results = [
            [0, 9, 4, 2, 6, 1, 3, 5, 7, 8],
            [1, 9, 5, 3, 7, 0, 2, 4, 6, 8],
            [2, 9, 5, 0, 7, 1, 3, 4, 6, 8],
            [3, 9, 0, 6, 1, 2, 4, 5, 7, 8],
            [4, 9, 0, 2, 6, 1, 3, 5, 7, 8],
            [5, 0, 9, 2, 7, 1, 3, 4, 6, 8],
            [6, 0, 3, 9, 1, 2, 4, 5, 7, 8],
            [7, 0, 3, 5, 9, 1, 2, 4, 6, 8],
            [8, 0, 4, 2, 6, 1, 3, 5, 7, 9],
            [9, 0, 5, 2, 7, 1, 6, 3, 8, 4],
        ]
    """
    sub_index_gen = _farthest_first_indices(0, len(items), first=first)
    if max_num is not None:
        sub_index_gen = it.islice(sub_index_gen, max_num)
    if isinstance(items, dict):
        item_keys = list(items.keys())
        for idx in sub_index_gen:
            key = item_keys[idx]
            value = items[key]
            yield (key, value)
    else:
        for idx in sub_index_gen:
            yield items[idx]


def _farthest_first_indices(start, stop, first=None):
    """
    Given a ordered list of items, incrementally yield indexes such that each
    new index maximizes the distance to the closest previously chosen index.

    Args:
        start (int): The inclusive starting index (typically 0)
        stop (int): The exclusive maximum index (typically ``len(items)``)
        first (int | None): The index to start from.
            If unspecified the largest index is first.

    Yields:
        int: the next chosen index in the series

    Notes:
        * This is an instance of farthest-point traversal in 1D.

    References:
        .. [CSSE_167943] https://cs.stackexchange.com/questions/167943/is-this-knapsack-variant-named-studied-online-algorithm-for-farthest-from-pr
        .. [WikiFarthestFirst] https://en.wikipedia.org/wiki/Farthest-first_traversal

    CommandLine:
        xdoctest -m geowatch.utils.util_kwutil _farthest_first_indices

    Example:
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> total = 10
        >>> start, stop = 0, 10
        >>> gen = _farthest_first_indices(start, stop)
        >>> result = list(gen)
        >>> assert set(result) == set(range(start, stop))
        >>> print(result)
        [9, 0, 5, 2, 7, 1, 6, 3, 8, 4]

    Example:
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> from geowatch.utils.util_kwutil import _farthest_first_indices
        >>> from geowatch.utils.util_kwutil import _check_farthest_first_index_sequence
        >>> import ubelt as ub
        >>> start, stop = 0, 10
        >>> gen = _farthest_first_indices(start, stop)
        >>> results = []
        >>> for first in range(start, stop):
        >>>     result = list(_farthest_first_indices(start, stop, first=first))
        >>>     assert _check_farthest_first_index_sequence(result)
        >>>     results += [result]
        >>> print(f'results = {ub.urepr(results, nl=1)}')
        results = [
            [0, 9, 4, 2, 6, 1, 3, 5, 7, 8],
            [1, 9, 5, 3, 7, 0, 2, 4, 6, 8],
            [2, 9, 5, 0, 7, 1, 3, 4, 6, 8],
            [3, 9, 0, 6, 1, 2, 4, 5, 7, 8],
            [4, 9, 0, 2, 6, 1, 3, 5, 7, 8],
            [5, 0, 9, 2, 7, 1, 3, 4, 6, 8],
            [6, 0, 3, 9, 1, 2, 4, 5, 7, 8],
            [7, 0, 3, 5, 9, 1, 2, 4, 6, 8],
            [8, 0, 4, 2, 6, 1, 3, 5, 7, 9],
            [9, 0, 5, 2, 7, 1, 6, 3, 8, 4],
        ]
        >>> result = list(gen)
        >>> assert set(result) == set(range(start, stop))
        >>> print(result)

        print(list(_farthest_first_indices(0, 32)))

        start = 0
        for stop in range(100):
            for first in range(start, stop):
                result = list(_farthest_first_indices(start, stop, first=first))
                assert _check_farthest_first_index_sequence(result)
    """
    import ubelt as ub
    if first is None or first == stop - 1:
        # Standard case
        if start < stop:
            yield stop - 1
            yield from _farthest_from_previous_helper(start, stop - 1)
    else:
        yield from _farthest_first_indices_generalized(start, stop, first)

    # Broken
    if 0:
        # Case where first is specified
        yield first
        # In this case, the left and right are not always balanced, so we generate
        # from whatever one is further until they balance. We have to include
        # the "first" element in both left and right generators to ensure
        # are generating elments far away from it, so we have to filter them out.
        left_gen = _farthest_first_indices(start, first + 1)
        right_gen = _farthest_first_indices(first, stop)

        DEBUG = 1
        if DEBUG:
            from itertools import tee
            right_gen, right_gen1 = tee(right_gen)
            left_gen, left_gen1 = tee(left_gen)
            print(f'start = {ub.urepr(start, nl=1)}')
            print(f'stop={stop}')
            print(f'first = {ub.urepr(first, nl=1)}')
            print('setup left: ' + str(list(left_gen1)))
            print('setup right: ' + str(list(right_gen1)))

        # the first item in the left gen will be "first", which we already yeilded
        # pop it off and ignore it.
        next(left_gen)

        # As long as first isn't the same as start or stop, then next will not raise StopIteration here.
        next_left = next(left_gen, None)
        next_right = next(right_gen, None)

        """
        Observation:
            the "best" distance halfs every time we take an item past the first.

            def best_dist(sequence):
                for idx, item in enumerate(sequence):
                    if idx == 0:
                        yield len(sequence)
                    else:
                        yield min([abs(p - item) for p in sequence[:idx]])
            sequence = list(_farthest_first_indices(0, 32))
            distance = list(best_dist(sequence))
            print(ub.urepr(ub.dzip(sequence, distance), align=':'))
            print(ub.urepr(ub.dict_hist(distance)))

            If the previous distance id d:
                then the next distance can only be on of:
                    d
                    d - 1
                    (d // 2)
                    (d // 2) + 1

            sequence: [31, 0, 16, 8, 24, 4, 20, 12, 28, 2, 18, 10, 26, 6, 22, 14, 30, 1, 17, 9, 25, 5, 21, 13, 29, 3, 19, 11, 27, 7, 23, 15]
            bestdist:  ∞  31, 15, 8, 7, 4, 4, 4, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]

            The 0th selection is basically free and doesn't have a value.
            the 1st will always have a distance of length of the sequence.
            the 2nd selection will roughly half the distance
            the 3rd will roughly half the distance again
            the 4th will be within 1 of the halfed distance
            the 5th will roughly half the distance again and there will be 4 of this magnitude
            the 9th will be the next halving and have 8 items at the magnitude
            The 17th item will be the next halving
            The 33th will be the next half

            so the pattern is the i-th item's distance is within at most this
            number of halfings:

                from numpy import log2, ceil

                def num_halfings(i):
                    if i == 0:
                        return 0
                    else:
                        return ceil(log2(i))

                def max_distance_at_step(i, total):
                    if i == 0:
                        return total
                    return total / (2 ** ceil(log2(i)))

                for total in range(2048):
                    bound2 = np.array([max_distance_at_step(i, total) for i in range(total)], dtype=int)
                    halvings = np.array([num_halfings(i) for i in range(total)], dtype=int)
                    bound = np.floor(total / 2 ** halvings).astype(int)
                    sequence = np.array(list(_farthest_first_indices(0, total)))
                    distance = np.array(list(best_dist(sequence)))
                    print(sequence)
                    print(distance)
                    print(bound)
                    np.array(distance <= bound)
                    tightness = distance - bound
                    assert np.all(np.abs(tightness) <= 1)

               so if we have a first element,
               a left sequence an an optimal right sequence, we know the
               size of the left and right sequence.

               without loss of generality assume the right sequence is larger.

               the right size has size R
               and the left side has size L
               in the case R > L

               in either sequence, the distance will roughly half with the sequence:

                   0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 5, ...  # https://oeis.org/A029837


               say
               start=0
               stop=81
               first=27

               result = list(_farthest_first_indices(start, stop, first=first))

               so

               R = 53
               L = 27

               The max distance in each case will be...

               [int(max_distance_at_step(i, R)) for i in range(1, 4)]
               [int(max_distance_at_step(i, L)) for i in range(1, 4)]


        """

        if next_left is None:
            if next_right is not None:
                # FIXME; something wrong here in the case where first = stop
                yield next_right
                # Ignore the next item in next right, it "first"
                next(right_gen, None)
                yield from right_gen
            return
        elif next_right is None:
            yield next_left
            yield from left_gen
            return

        left_dist = abs(next_left - first)
        right_dist = abs(next_right - first)

        left_size = first - start
        right_size = stop - first - 1
        print(f'left_size={left_size}')
        print(f'right_size={right_size}')

        # The second item in the right gen will be "first", so we pop it off
        # to ignore it.
        next(right_gen, None)

        DEBUG = 1
        if DEBUG:
            from itertools import tee
            right_gen, right_gen1 = tee(right_gen)
            left_gen, left_gen1 = tee(left_gen)
            print(f'start left: {next_left}, ' + str(list(left_gen1)))
            print(f'start right: {next_right}, ' + str(list(right_gen1)))
            #### FFF this doesn't work..

        while True:
            if left_dist == right_dist:
                # At this point the distance to first heuristic breaks in the
                # case where start=0, stop=10, first=2. I think it might be the
                # case that when the heuristic breaks we can just interleave
                # the remaining iterables, but I'm not sure if this holds.
                total = (stop - start)
                if DEBUG:
                    print('At a point where left and right dist are the same')
                    right_gen, right_gen1 = tee(right_gen)
                    left_gen, left_gen1 = tee(left_gen)
                    print(f'curr left: {next_left}, ' + str(list(left_gen1)))
                    print(f'curr right: {next_right}, ' + str(list(right_gen1)))
                    print(f'total = {ub.urepr(total, nl=1)}')

                # if total % 2 == 0:
                pairgen = it.zip_longest(left_gen, right_gen)
                flatgen = it.chain([next_left], [next_right], *pairgen)
                filtgen = filter(lambda x: x is not None, flatgen)
                yield from filtgen
                # else:
                #     pairgen = it.zip_longest(right_gen, left_gen)
                #     flatgen = it.chain([next_right], [next_left], *pairgen)
                #     filtgen = filter(lambda x: x is not None, flatgen)
                #     yield from filtgen
                return
            if left_dist > right_dist:
                if DEBUG:
                    print(f'chose left {next_left=} {next_right=} {left_dist=} {right_dist=}')
                yield next_left
                next_left = next(left_gen, None)
                if next_left is None:
                    # No more data on the left side, finish the right side
                    yield next_right
                    yield from right_gen
                    break
                left_dist = abs(next_left - first)
            else:
                if DEBUG:
                    print(f'chose right {next_left=} {next_right=} {left_dist=} {right_dist=}')
                yield next_right
                next_right = next(right_gen, None)
                if next_right is None:
                    # No more data on the right side, finish the left side
                    yield next_left
                    yield from left_gen
                    break
                right_dist = abs(next_right - first)


def _check_farthest_first_index_sequence(sequence):
    """
    Checks that the sequence of indices satisfies the farthest-first objective.

    For each item in the sequence, it should be the further from any seen value
    than any unseen value.

    Example:
        >>> # There are multiple ways a sequence can be valie
        >>> assert _check_farthest_first_index_sequence([3, 0, 2, 1])
        >>> assert _check_farthest_first_index_sequence([0, 3, 2, 1])
        >>> assert not _check_farthest_first_index_sequence([0, 2, 3, 1])
    """
    for idx, index in enumerate(sequence):
        seen = sequence[:idx]
        unseen = sequence[idx + 1:]
        # Look at the distances from the chosen item to all seen items
        this_dist_to_seens = {prev: abs(index - prev) for prev in seen}
        # Look at the distances from the unchosen item to all seen items
        other_dist_to_seens = {
            other: {prev: abs(other - prev) for prev in seen}
            for other in unseen
        }
        import ubelt as ub
        if seen and unseen:
            # The minimum distance from this to the seen items should be greater than
            this_min_dist_to_seen_value = min(this_dist_to_seens.values())
            # The minimum distance from any unseen item to the seen items
            other_min_dist_to_seen_value = max([min(o.values()) for o in other_dist_to_seens.values()])
            next_best = ub.argmax({k: min(o.values()) for k, o in other_dist_to_seens.items()})
            if this_min_dist_to_seen_value < other_min_dist_to_seen_value:
                print('----')
                print(f'index={index}')
                print(f'seen = {ub.urepr(seen, nl=0)}')
                print(f'unseen = {ub.urepr(unseen, nl=0)}')
                print(f'this_dist_to_seens = {ub.urepr(this_dist_to_seens, nl=1)}')
                print(f'other_dist_to_seens = {ub.urepr(other_dist_to_seens, nl=1)}')
                print(f'this_min_dist_to_seen_value={this_min_dist_to_seen_value}')
                print(f'other_min_dist_to_seen_value={other_min_dist_to_seen_value}')
                print(f'We chose the index {index}, but we could have chosen {next_best}')
                print(f'sequence = {ub.urepr(sequence, nl=0)}')
                return False

    return True


def _farthest_from_previous_helper(start, stop):
    if start < stop:
        low_mid: int = (start + stop) // 2
        high_mid: int = (start + stop + 1) // 2

        left_gen = _farthest_from_previous_helper(start, low_mid)
        right_gen = _farthest_from_previous_helper(high_mid, stop)

        pairgen = it.zip_longest(left_gen, right_gen)
        flatgen = it.chain.from_iterable(pairgen)
        filtgen = filter(lambda x: x is not None, flatgen)
        yield from filtgen
        if low_mid < high_mid:
            yield low_mid


def _farthest_first_indices_generalized(start, stop, first=None, distance_fn=None):
    """
    Yield indices in a farthest-first traversal order for a 1D uniform grid [start, stop),
    with an optional user-specified starting index.

    This implementation uses a heap to maintain for each candidate the minimum distance
    to any selected index, so that we can always pick the candidate that maximizes that distance.

    Args:
        start (int): inclusive start index.
        stop (int): exclusive stop index.
        first (int, optional): the starting index (must be in [start, stop)). Defaults to stop - 1.
        distance_fn (callable, optional): A function that computes distance between two indices.
            Defaults to absolute difference (L1 norm).

    Yields:
        int: the next index in the farthest-first order.

    References:
        https://chatgpt.com/c/67a37e86-f89c-8013-8ef8-400c209de0e0

    TODO:
        - [ ] Allow the user to specify a custom distance metric. The logic
              should still work because we only care about the maximum distance
              of unselected elements to their closest selected index.

    Example:
        >>> from geowatch.utils.util_kwutil import *  # NOQA
        >>> import ubelt as ub
        >>> total = 10
        >>> start, stop = 0, 10
        >>> results = []
        >>> for first in range(start, stop):
        >>>     results += [list(_farthest_first_indices_generalized(start, stop, first=first))]
        >>> print(f'results = {ub.urepr(results, nl=1)}')
        results = [
            [0, 9, 4, 2, 6, 1, 3, 5, 7, 8],
            [1, 9, 5, 3, 7, 0, 2, 4, 6, 8],
            [2, 9, 5, 0, 7, 1, 3, 4, 6, 8],
            [3, 9, 0, 6, 1, 2, 4, 5, 7, 8],
            [4, 9, 0, 2, 6, 1, 3, 5, 7, 8],
            [5, 0, 9, 2, 7, 1, 3, 4, 6, 8],
            [6, 0, 3, 9, 1, 2, 4, 5, 7, 8],
            [7, 0, 3, 5, 9, 1, 2, 4, 6, 8],
            [8, 0, 4, 2, 6, 1, 3, 5, 7, 9],
            [9, 0, 4, 2, 6, 1, 3, 5, 7, 8],
        ]
        >>> # This distance function just reverses the metric, so we find the
        >>> # closest item instead of the furthest. It is a bit contrived, but
        >>> # demos how a specific distance function can modify the results.
        >>> results = []
        >>> def distance_fn(i, j):
        >>>     return -abs(i - j)
        >>> for first in range(start, stop):
        >>>     results += [list(_farthest_first_indices_generalized(start, stop, first=first, distance_fn=distance_fn))]
        >>> print(f'results = {ub.urepr(results, nl=1)}')
        results = [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            [1, 0, 2, 3, 4, 5, 6, 7, 8, 9],
            [2, 1, 0, 3, 4, 5, 6, 7, 8, 9],
            [3, 2, 1, 0, 4, 5, 6, 7, 8, 9],
            [4, 3, 2, 1, 0, 5, 6, 7, 8, 9],
            [5, 4, 3, 2, 1, 0, 6, 7, 8, 9],
            [6, 5, 4, 3, 2, 1, 0, 7, 8, 9],
            [7, 6, 5, 4, 3, 2, 1, 0, 8, 9],
            [8, 7, 6, 5, 4, 3, 2, 1, 0, 9],
            [9, 8, 7, 6, 5, 4, 3, 2, 1, 0],
        ]
    """
    # Note: using sortedcontainers might have a more clear implementation this
    # is probably still O(n**2), but I think we must be able to do better.  not
    # going to think about it too hard, this code path is not very likely to be
    # needed, as the simple farther first implementation is what is mainly
    # needed.
    import heapq
    if first is None:
        first = stop - 1
    if not (start <= first < stop):
        raise ValueError("first must be in the range [start, stop)")

    if distance_fn is None:
        distance_fn = lambda a, b: abs(a - b)  # NOQA

    # Set of selected indices (kept sorted for convenience).
    selected = [first]
    yield first

    # For each candidate (not yet selected), store its best known distance to a selected index.
    best_dist = {
        i: distance_fn(i, first)
        for i in range(start, stop)
        if i != first
    }
    # Maintain a heap keyed by negative distance (since heapq is a min-heap).
    heap = []
    for i, d in best_dist.items():
        heapq.heappush(heap, (-d, i))

    # Greedily select the candidate with the maximum current distance.
    while heap:
        neg_d, candidate = heapq.heappop(heap)
        # If candidate was already selected, skip it.
        if candidate not in best_dist:
            continue
        # If the min distance in the heap is wrong, then the item is outdated,
        # and we need to skip it.
        current_d = best_dist[candidate]
        if -neg_d != current_d:
            # The stored distance is outdated, pop it and skip, the correct
            # value will exist in the queue later on.
            continue

        # Candidate is the farthest from any selected point.
        yield candidate
        selected.append(candidate)
        # Remove candidate from best_dist so it isn’t chosen again.
        best_dist.pop(candidate)

        # Update best distances for all remaining candidates.
        for other, old_d in best_dist.items():
            new_d = distance_fn(other, candidate)
            if new_d < old_d:
                best_dist[other] = new_d
                heapq.heappush(heap, (-new_d, other))



import ubelt as ub
import pytest
import kwcoco
from geowatch.cli import coco_align
from geowatch.geoannots import geomodels


def test_coco_align_with_empty_site_summary():
    coco_dset = kwcoco.CocoDataset()

    # Create arguments to the script
    dpath = ub.Path.appdir('geowatch/tests/align_bundle_empty1').delete().ensuredir()
    region_dpath = (dpath / 'region_models').ensuredir()

    # Write a region model with no sites, and run align in "site-summary" mode.
    region = geomodels.RegionModel.random(num_sites=0)
    region_fpath = region_dpath / 'region.geojson'
    region.dump(region_fpath)
    dst_dpath = (dpath / 'dst_bundle').ensuredir()
    dst_fpath = dst_dpath / 'dst.kwcoco.zip'

    kw = {
        'src': coco_dset,
        'dst': dst_fpath,
        'keep': 'img',
        'regions': region_fpath,
        'workers': 2,
        'aux_workers': 2,
        'convexify_regions': True,
        'convexify_regions': True,
        'num_start_frames': 3,
        'num_end_frames': 3,
        'site_summary': True,
        'minimum_size': '128x128@10GSD',
    }
    coco_align.main(cmdline=False, **kw)
    assert dst_fpath.exists()
    dst = kwcoco.CocoDataset(dst_fpath)
    assert dst.n_images == 0

    # Test with empty region policy as raise
    kw['empty_region_policy'] = 'raise'
    with pytest.raises(ValueError):
        coco_align.main(cmdline=False, **kw)


def test_coco_align_with_empty_inputs_coco():
    import kwcoco
    from geowatch.cli import coco_align
    from geowatch.geoannots import geomodels
    coco_dset = kwcoco.CocoDataset()

    # Create arguments to the script
    dpath = ub.Path.appdir('geowatch/tests/align_bundle_empty2').delete().ensuredir()
    region_dpath = (dpath / 'region_models').ensuredir()

    region = geomodels.RegionModel.random()
    region_fpath = region_dpath / 'region.geojson'
    region.dump(region_fpath)
    dst_dpath = (dpath / 'dst_bundle').ensuredir()
    dst_fpath = dst_dpath / 'dst.kwcoco.zip'

    kw = {
        'src': coco_dset,
        'dst': dst_fpath,
        'keep': 'img',
        'regions': region_fpath,
        'workers': 2,
        'aux_workers': 2,
        'convexify_regions': True,
        'convexify_regions': True,
        'num_start_frames': 3,
        'num_end_frames': 3,
        'site_summary': True,
        'minimum_size': '128x128@10GSD',
    }
    coco_align.main(cmdline=False, **kw)
    assert dst_fpath.exists()
    dst = kwcoco.CocoDataset(dst_fpath)
    assert dst.n_images == 0
